from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_parse_list_with_one_item():
    source = """
    * This is a list with one element
    """

    expected = [
        {
            "type": "list",
            "ordered": False,
            "main_node": True,
            "items": [
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a list with one element",
                            }
                        ],
                    },
                }
            ],
        }
    ]

    _test(source, expected)


def test_parse_list_with_multiple_items():
    source = """
    * Item 1
    * Item 2
    """

    expected = [
        {
            "type": "list",
            "ordered": False,
            "main_node": True,
            "items": [
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "Item 1",
                            }
                        ],
                    },
                },
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "Item 2",
                            }
                        ],
                    },
                },
            ],
        }
    ]

    _test(source, expected)


def test_parse_list_with_multiple_levels():
    source = """
    * Item 1
    ** Item 1.1
    * Item 2
    """

    expected = [
        {
            "type": "list",
            "ordered": False,
            "main_node": True,
            "items": [
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "Item 1",
                            },
                            {
                                "type": "list",
                                "ordered": False,
                                "main_node": False,
                                "items": [
                                    {
                                        "type": "list_item",
                                        "level": 2,
                                        "content": {
                                            "type": "sentence",
                                            "content": [
                                                {
                                                    "type": "text",
                                                    "value": "Item 1.1",
                                                }
                                            ],
                                        },
                                    }
                                ],
                            },
                        ],
                    },
                },
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "Item 2",
                            }
                        ],
                    },
                },
            ],
        }
    ]

    _test(source, expected)


def test_parse_numbered_list():
    source = """
    # Item 1
    ## Item 1.1
    # Item 2
    """

    expected = [
        {
            "type": "list",
            "ordered": True,
            "main_node": True,
            "items": [
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "Item 1",
                            },
                            {
                                "type": "list",
                                "ordered": True,
                                "main_node": False,
                                "items": [
                                    {
                                        "type": "list_item",
                                        "level": 2,
                                        "content": {
                                            "type": "sentence",
                                            "content": [
                                                {
                                                    "type": "text",
                                                    "value": "Item 1.1",
                                                }
                                            ],
                                        },
                                    }
                                ],
                            },
                        ],
                    },
                },
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "Item 2",
                            }
                        ],
                    },
                },
            ],
        }
    ]

    _test(source, expected)


def test_parse_mixed_list():
    source = """
    * Item 1
    ## Item 1.1
    * Item 2
    """

    expected = [
        {
            "type": "list",
            "ordered": False,
            "main_node": True,
            "items": [
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "Item 1",
                            },
                            {
                                "type": "list",
                                "ordered": True,
                                "main_node": False,
                                "items": [
                                    {
                                        "type": "list_item",
                                        "level": 2,
                                        "content": {
                                            "type": "sentence",
                                            "content": [
                                                {
                                                    "type": "text",
                                                    "value": "Item 1.1",
                                                }
                                            ],
                                        },
                                    }
                                ],
                            },
                        ],
                    },
                },
                {
                    "type": "list_item",
                    "level": 1,
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "Item 2",
                            }
                        ],
                    },
                },
            ],
        }
    ]

    _test(source, expected)



# -*- coding: utf-8 -*-

import argparse
import logging
import sys

import yaml

from mau import __version__, Mau, ConfigurationError, ParserError

__author__ = "Leonardo Giordani"
__copyright__ = "Leonardo Giordani"
__license__ = "mit"

_logger = logging.getLogger(__name__)

# This is a convenience dictionary that collects the output formats
# that Mau supports and some matadata
FORMATS = {
    "html": {"extension": ".html"},
    "asciidoctor": {"extension": ".adoc"},
    "markua": {"extension": ".md"},
}


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input-file",
        action="store",
        required=True,
        help="Input file",
    )

    parser.add_argument(
        "-o",
        "--output-file",
        action="store",
        required=False,
        help="Optional output file",
    )

    parser.add_argument(
        "-c",
        "--config-file",
        action="store",
        required=False,
        help="Optional YAML config file",
    )

    parser.add_argument(
        "-f",
        "--format",
        action="store",
        required=False,
        choices=FORMATS.keys(),
        help="Output format",
    )

    parser.add_argument(
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )

    parser.add_argument(
        "--debug",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )

    parser.add_argument(
        "--version", action="version", version=f"Mau version {__version__}"
    )

    return parser.parse_args()


def setup_logging(loglevel):
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    # Get arguments and logging set up
    args = parse_args(args)
    setup_logging(args.loglevel)

    # Read the input file
    with open(args.input_file, "r") as f:
        text = f.read()

    # Load the YAML configuration file into a dictionary
    config = {}
    if args.config_file:
        with open(args.config_file, "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)

    # The directory that contains custom templates.
    templates_directory = config.get("templates_directory", None)

    # If a target format is specified on the command line
    # override the one already set in the config file
    # or set it if it isn't.
    if args.format is not None:
        config["target_format"] = args.format

    # This extracts the custom templates defined
    # in the config file
    custom_templates = config.get("custom_templates", None)

    # The Mau object configured with what we figured out above.
    mau = Mau(
        config,
        None,
        custom_templates=custom_templates,
        templates_directory=templates_directory,
        full_document=True,
    )

    try:
        # Process the input text.
        output = mau.process(text)
    except ConfigurationError as e:
        print(f"Configuration error: {e}")
        sys.exit(1)
    except ParserError as e:
        print(str(e))
        print()
        print(f"Line {e.context['line']}, column {e.context['column']}")
        for i in e.context["text"]:
            print(i)
        sys.exit(1)

    # Find out the name of the output file
    output_extension = FORMATS[config["target_format"]]["extension"]
    output_file = args.output_file or args.input_file.replace(".mau", output_extension)

    # Write the output
    with open(output_file, "w") as f:
        text = f.write(output)


def run():
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



import pytest

from unittest.mock import patch

from mau.parsers.nodes import DocumentNode
from mau.parsers.main_parser import MainParser
from mau.visitors.asciidoctor_visitor import AsciidoctorVisitor

from tests.helpers import (
    init_parser_factory,
    ast_test_factory,
    init_ast_factory,
    dedent,
    listasdict,
    visitlist_factory,
)

init_parser = init_parser_factory(MainParser)
init_ast = init_ast_factory(MainParser)

_test = ast_test_factory(MainParser, AsciidoctorVisitor)

visitlist = visitlist_factory(AsciidoctorVisitor)


def test_text():
    v = AsciidoctorVisitor()

    output = v.visit(
        {
            "type": "text",
            "value": "This is text",
        }
    )

    assert output == "This is text"


def test_sentence():
    v = AsciidoctorVisitor()

    output = v.visit(
        {"type": "sentence", "content": [{"type": "text", "value": "This is text"}]}
    )

    assert output == "This is text"


def test_paragraph():
    source = "This is text"

    expected = ["This is text\n"]

    _test(source, expected)


def test_horizontal_rule():
    source = "---"

    expected = ["---\n"]

    _test(source, expected)


def test_style_underscore():
    source = "_This is text_"

    expected = ["_This is text_\n"]

    _test(source, expected)


def test_text_that_contains_styles():
    source = "This _is_ text"

    expected = ["This _is_ text\n"]

    _test(source, expected)


def test_style_star():
    source = "*This is text*"

    expected = ["*This is text*\n"]

    _test(source, expected)


def test_style_verbatim():
    source = "`This is text`"

    expected = ["`This is text`\n"]

    _test(source, expected)


def test_class():
    source = "[classname]#text with that class#"

    expected = ["[.classname]#text with that class#\n"]

    _test(source, expected)


def test_link():
    source = '[link](https://somedomain.org/the/path,"link text")'

    expected = ["https://somedomain.org/the/path[link text]\n"]

    _test(source, expected)


def test_link_target_equal_text():
    source = "https://somedomain.org/the/path"

    expected = ["https://somedomain.org/the/path\n"]

    _test(source, expected)


def test_inline_image():
    source = '[image](/the/path.jpg,"alt text")'

    expected = ["image::/the/path.jpg[alt text]\n"]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_header_level_3(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = "=== Title of the section"

    expected = ["=== Title of the section\n"]

    _test(source, expected)


def test_quote():
    source = dedent(
        """
        [quote,"Star Wars, 1977"]
        ----
        Learn about the Force, Luke.
        ----
        """
    )

    expected = [
        '[quote, "Star Wars, 1977"]\n____\nLearn about the Force, Luke.\n____\n'
    ]

    _test(source, expected)


def test_admonition():
    source = dedent(
        """
        [admonition,warning,anicon,alabel]
        ----
        This is my test admonition
        ----
        """
    )

    expected = ["[WARNING.anicon]\n====\nThis is my test admonition\n====\n"]

    _test(source, expected)


def test_admonition_not_known():
    ast = init_ast(
        dedent(
            """
            [admonition,something,anicon,alabel]
            ----
            This is my test admonition
            ----
            """
        )
    )

    with pytest.raises(ValueError):
        visitlist(ast)


@patch("mau.parsers.main_parser.header_anchor")
def test_block(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = dedent(
        """
        ++++
        = A block

        This contains headers, paragraphs and blocks

        ++++
        """
    )

    expected = [
        dedent(
            """
            --
            = A block

            This contains headers, paragraphs and blocks

            --
            """
        )
    ]

    _test(source, expected)


def test_footnote():
    p = init_parser(
        dedent(
            """
            This is a sentence[footnote](with a note)
            """
        )
    )

    p.parse()
    ast = listasdict(p.nodes)

    result = visitlist(ast, footnotes=p.footnotes)

    assert result == ["This is a sentencefootnote:[with a note]\n"]


def test_unordered_list():
    source = dedent(
        """
        * Item 1
        * Item 2
        ** Item 2.1
        *** Item 2.1.1
        * Item 3
        """
    )

    expected = ["\n* Item 1\n* Item 2\n** Item 2.1\n*** Item 2.1.1\n* Item 3\n"]

    _test(source, expected)


def test_ordered_list():
    source = dedent(
        """
        # Item 1
        # Item 2
        ## Item 2.1
        ### Item 2.1.1
        # Item 3
        """
    )

    expected = ["\n. Item 1\n. Item 2\n.. Item 2.1\n... Item 2.1.1\n. Item 3\n"]

    _test(source, expected)


def test_document():
    parser = init_parser(
        dedent(
            """
            This is text
            """
        )
    )
    parser.parse()

    document = DocumentNode(parser.nodes)
    node = document.asdict()

    result = visitlist([node])

    assert result == ["This is text\n"]


def test_source():
    source = dedent(
        """
        [myblock, engine=source, language=somelang]
        ----
        Some {} source a:=5 code
        ----
        """
    )

    expected = ["[source,somelang]\n----\nSome {} source a:=5 code\n----\n"]

    _test(source, expected)


def test_source_title():
    source = dedent(
        """
        .title
        [myblock, engine=source, language=somelang]
        ----
        Some {} source a:=5 code
        ----
        """
    )

    expected = [".title\n[source,somelang]\n----\nSome {} source a:=5 code\n----\n"]

    _test(source, expected)


def test_source_callouts():
    source = dedent(
        """
        [myblock, engine=source, language=somelang]
        ----
        Some {} source a:=5 code:1:
        ----
        1: Callout
        """
    )

    expected = [
        "[source,somelang]\n----\nSome {} source a:=5 code <1>\n----\n<1> Callout\n"
    ]

    _test(source, expected)


def test_content_image():
    source = dedent(
        """
        << image:/path/to/it.jpg
        """
    )

    expected = ["image::/path/to/it.jpg[]\n"]

    _test(source, expected)


def test_content_image_with_title_and_alt_text():
    source = dedent(
        """
        . Title
        [alt_text="Some text"]
        << image:/path/to/it.jpg
        """
    )

    expected = [".Title\nimage::/path/to/it.jpg[Some text]\n"]

    _test(source, expected)


def test_content_image_asciidoctor_classes():
    source = dedent(
        """
        . Title
        [alt_text="Some text", asciidoctor_classes=".text-center"]
        << image:/path/to/it.jpg
        """
    )

    expected = ["[.text-center]\n.Title\nimage::/path/to/it.jpg[Some text]\n"]

    _test(source, expected)



import pytest

from unittest.mock import patch

from mau.lexers.base_lexer import (
    Token,
    TokenTypes,
    EOL,
    EOF,
    Text,
    Literal,
)
from mau.lexers.base_lexer import BaseLexer
from mau.parsers.base_parser import BaseParser, ParserError, ExpectedError, TokenError


from tests.helpers import init_parser_factory

init_parser = init_parser_factory(BaseParser)


def test_initial_state():
    p = BaseParser()

    assert p.index == -1
    with pytest.raises(ValueError):
        p.current_token


def test_load():
    p = init_parser("\n")

    assert p.index == -1
    with pytest.raises(ValueError):
        p.current_token


def test_get_token_without_load():
    p = BaseParser()

    assert p.get_token() == EOF


def test_get_token():
    p = init_parser("\n")

    assert p.get_token() == EOL
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_put_token():
    p = init_parser("\n")

    assert p.get_token() == EOL
    p.put_token(Literal("*"))
    assert p.get_token() == Literal("*")
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_tokens_contain_position():
    p = init_parser("\n")

    assert p.get_token().position == (0, 0)


def test_token_position_is_updated():
    p = init_parser("\n")

    assert p.get_token() == EOL
    assert p.get_token() == EOL
    assert p.get_token().position == (2, 0)


def test_get_token_after_eof():
    p = init_parser("\n")

    assert p.index == -1
    assert p.get_token() == EOL
    assert p.index == 0
    assert p.get_token() == EOL
    assert p.index == 1
    assert p.get_token() == EOF
    assert p.index == 2
    assert p.get_token() == EOF
    assert p.index == 3
    assert p.get_token() == EOF
    assert p.index == 3


def test_get_token_sets_current_token():
    p = init_parser("\n")

    assert p.get_token() == EOL
    assert p.current_token == EOL

    assert p.get_token() == EOL
    assert p.get_token() == EOF
    assert p.current_token == EOF


def test_get_token_accepts_type():
    p = init_parser("\n")

    assert p.get_token(TokenTypes.EOL) == EOL
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_get_token_accepts_type_and_value():
    p = init_parser("\n")

    assert p.get_token(TokenTypes.EOL, None) == EOL
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_get_token_checks_type():
    p = init_parser("\n")

    with pytest.raises(TokenError):
        p.get_token(TokenTypes.EOF)

    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_get_token_checks_type_and_value():
    p = init_parser("\n")

    with pytest.raises(TokenError):
        p.get_token(TokenTypes.EOL, "foobar")

    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_get_token_accepts_check_function():
    p = init_parser("\n")

    with pytest.raises(TokenError):
        p.get_token(TokenTypes.EOF, check=lambda x: x is not None)

    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_check_current_token():
    p = init_parser("\n")
    p.get_token()

    assert p.check_current_token(TokenTypes.EOL) == EOL

    # check_current_token doesn't advance the index
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_check_current_token_raises_exception():
    p = init_parser("\n")
    p.get_token()

    with pytest.raises(TokenError):
        p.check_current_token(TokenTypes.EOF)

    # check_current_token doesn't advance the index
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_check_current_token_value():
    p = init_parser("\n")
    p.get_token()

    assert p.check_current_token(TokenTypes.EOL, None) == EOL

    # check_current_token doesn't advance the index
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_check_current_token_with_function():
    p = init_parser("\n")
    p.get_token()

    assert p.check_current_token(TokenTypes.EOL, check=lambda x: x is None) == Token(
        TokenTypes.EOL
    )

    # check_current_token doesn't advance the index
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_peek_token():
    p = init_parser("\n")

    assert p.peek_token() == EOL
    assert p.get_token() == EOL
    assert p.get_token() == EOL
    assert p.peek_token() == EOF
    assert p.get_token() == EOF


def test_base_lexer_can_peek_more_than_available():
    p = init_parser("\n")

    assert p.get_token() == EOL
    assert p.get_token() == EOL
    assert p.get_token() == EOF
    assert p.peek_token() == EOF


def test_peek_token_checks_type():
    p = init_parser("\n")

    p.get_token()  # Get the first EOL
    p.get_token()  # Get the second EOL

    assert p.peek_token(TokenTypes.EOF) == EOF
    assert p.get_token() == EOF


def test_base_lexer_check_next_token_raises_exception():
    p = init_parser("\n")

    p.get_token()  # Get the first EOL
    p.get_token()  # Get the second EOL

    with pytest.raises(TokenError):
        p.peek_token(TokenTypes.EOL)

    # failed attempt doesn't advance the index
    assert p.get_token() == EOF


def test_peek_token_checks_type_and_value():
    p = init_parser("\n")

    with pytest.raises(TokenError):
        p.peek_token(TokenTypes.EOL, "foobar")

    assert p.get_token() == EOL
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_peek_token_accepts_check_function():
    p = init_parser("\n")

    with pytest.raises(TokenError):
        p.peek_token(TokenTypes.EOL, check=lambda x: x is not None)

    assert p.get_token() == EOL
    assert p.get_token() == EOL
    assert p.get_token() == EOF


def test_context_manager():
    p = init_parser("\n")

    with p:
        assert p.get_token() == EOL


def test_do_not_restore_status_if_no_error():
    p = init_parser("\n")

    with p:
        assert p.get_token(TokenTypes.EOL)
        assert p.get_token(TokenTypes.EOL)

    assert p.get_token(TokenTypes.EOF)


def test_restore_status_if_token_error():
    p = init_parser("\n")

    with p:
        assert p.get_token(TokenTypes.EOL)
        assert p.get_token(TokenTypes.EOL)
        assert p.get_token(TokenTypes.EOL)

    assert p.get_token(TokenTypes.EOL)
    assert p.get_token(TokenTypes.EOL)
    assert p.get_token(TokenTypes.EOF)


def test_context_manager_leaves_exceptions_untouched():
    p = init_parser("\n")

    with pytest.raises(ValueError):
        with p:
            raise ValueError


def test_token_error_can_be_raised():
    p = init_parser("\n")

    with p:
        assert p.get_token(TokenTypes.EOL)
        raise TokenError

    assert p.get_token(TokenTypes.EOL)
    assert p.get_token(TokenTypes.EOL)
    assert p.get_token(TokenTypes.EOF)


def test_context_manager_nested():
    p = init_parser("\n")

    with p:
        p.get_token()
        with p:
            p.get_token(TokenTypes.EOF)

        p.get_token(TokenTypes.EOF)

    assert p.get_token(TokenTypes.EOL)
    assert p.get_token(TokenTypes.EOL)
    assert p.get_token(TokenTypes.EOF)


def test_context_manager_nested_with_success():
    p = init_parser("\n")

    with p:
        p.get_token()
        with p:
            p.get_token()
            # This succeeds
            p.get_token(TokenTypes.EOF)

        p.get_token(TokenTypes.EOL)

    assert p.get_token(TokenTypes.EOL)
    assert p.get_token(TokenTypes.EOL)
    assert p.get_token(TokenTypes.EOF)


def test_peek_token_is():
    p = init_parser("\n")

    p.get_token()  # Get the first EOL
    p.get_token()  # Get the second EOL

    assert p.peek_token_is(TokenTypes.EOL) is False
    assert p.peek_token_is(TokenTypes.EOF) is True


def test_force_token():
    p = init_parser("\n")

    assert p.force_token(TokenTypes.EOL) == EOL
    assert p.force_token(TokenTypes.EOL) == EOL

    with pytest.raises(ExpectedError):
        p.force_token(TokenTypes.EOL)


def test_collect():
    p = init_parser("\n")

    tokens = p.collect([EOF])

    assert tokens == [EOL, EOL]

    tokens = p.collect([EOF])

    assert tokens == []


def test_collect_join():
    p = init_parser("\n")
    p.tokens = [
        Text("Some te"),
        Text("xt that will be joined"),
        Literal("!"),
        EOL,
        EOL,
    ]

    expected = "Some text that will be joined!"

    assert p.collect_join([EOF]) == expected


def test_collect_join_with_different_joiner():
    p = init_parser("\n")
    p.tokens = [
        Text("De-do-do"),
        Text("do"),
        EOL,
        EOL,
    ]

    expected = "De-do-do-do"

    assert p.collect_join([EOF], "-") == expected


@patch.object(BaseLexer, "context")
def test_error(mock_context):
    p = init_parser("Some text")
    p.get_token()

    with pytest.raises(ParserError):
        p.error()

    assert mock_context.called


@patch.object(BaseLexer, "context")
def test_error_with_message(mock_context):
    p = init_parser("Some text")
    p.get_token()

    with pytest.raises(ParserError) as e:
        p.error("Just a message")

    assert mock_context.called
    assert str(e.value) == "Just a message"
    assert e.value.context == mock_context.return_value



from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_include_image_with_only_path():
    source = """
    << image:/path/to/it.jpg
    """

    expected = [
        {
            "type": "content_image",
            "uri": "/path/to/it.jpg",
            "alt_text": None,
            "title": None,
            "classes": None,
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_include_image_with_full_uri():
    source = """
    << image:/path/to/it.jpg
    """

    expected = [
        {
            "type": "content_image",
            "uri": "/path/to/it.jpg",
            "alt_text": None,
            "title": None,
            "classes": None,
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_include_image_with_http():
    source = """
    << image:https:///some.domain/path/to/it.jpg
    """

    expected = [
        {
            "type": "content_image",
            "uri": "https:///some.domain/path/to/it.jpg",
            "alt_text": None,
            "title": None,
            "classes": None,
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_include_image_with_alt_text():
    source = """
    [alt_text="This is a beautiful image"]
    << image:/path/to/it.jpg
    """

    expected = [
        {
            "type": "content_image",
            "uri": "/path/to/it.jpg",
            "alt_text": "This is a beautiful image",
            "title": None,
            "classes": None,
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_include_image_with_classes():
    source = """
    [classes="class1,class2"]
    << image:/path/to/it.jpg
    """

    expected = [
        {
            "type": "content_image",
            "uri": "/path/to/it.jpg",
            "alt_text": None,
            "title": None,
            "classes": ["class1", "class2"],
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_include_image_with_arguments():
    source = """
   [argument1=value1,argument2=value2]
   << image:/path/to/it.jpg
   """

    expected = [
        {
            "type": "content_image",
            "uri": "/path/to/it.jpg",
            "alt_text": None,
            "title": None,
            "classes": None,
            "kwargs": {"argument1": "value1", "argument2": "value2"},
        },
    ]

    _test(source, expected)


def test_include_image_with_title():
    source = """
    . A nice caption
    << image:/path/to/it.jpg
    """

    expected = [
        {
            "type": "content_image",
            "uri": "/path/to/it.jpg",
            "alt_text": None,
            "title": {
                "content": [{"type": "text", "value": "A nice caption"}],
                "type": "sentence",
            },
            "classes": None,
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_include_image_first_unnamed_attribute_is_alt_text():
    source = """
    ["Alternate text"]
    << image:/path/to/it.jpg
    """

    expected = [
        {
            "type": "content_image",
            "uri": "/path/to/it.jpg",
            "alt_text": "Alternate text",
            "title": None,
            "classes": None,
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_include_unknown_content():
    source = """
    . Some title
    [arg1,arg2,named1="value1",named2="value2"]
    << unknown:/path/to/it
    """

    expected = [
        {
            "type": "content",
            "uri": "/path/to/it",
            "title": {
                "content": [{"type": "text", "value": "Some title"}],
                "type": "sentence",
            },
            "args": ["arg1", "arg2"],
            "kwargs": {"named1": "value1", "named2": "value2"},
        },
    ]

    _test(source, expected)



from mau.parsers.main_parser import MainParser

from tests.helpers import listasdict, init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_parse_variable_definition_without_value_is_loaded_as_boolean():
    p = init_parser(":attr:")
    p.parse()

    assert listasdict(p.nodes) == []
    assert p.variables == {"mau": {}, "attr": True}


def test_parse_variable_definition_without_value_can_be_negated():
    p = init_parser(":!attr:")
    p.parse()

    assert listasdict(p.nodes) == []
    assert p.variables == {"mau": {}, "attr": False}


def test_parse_variable_definition_negative_flag_ignores_value():
    p = init_parser(":!attr:42")
    p.parse()

    assert listasdict(p.nodes) == []
    assert p.variables == {"mau": {}, "attr": False}


def test_parse_variable_definition_with_value_is_loaded():
    p = init_parser(":attr:42")
    p.parse()

    assert listasdict(p.nodes) == []
    assert p.variables == {"mau": {}, "attr": "42"}


def test_parse_variable_definition_value_can_be_any_text():
    p = init_parser(":attr:[footnote](http://some.domain/path)")
    p.parse()

    assert listasdict(p.nodes) == []
    assert p.variables == {"mau": {}, "attr": "[footnote](http://some.domain/path)"}


def test_parse_variable_definition_with_namespace():
    p = init_parser(":meta.attr:42")
    p.parse()

    assert listasdict(p.nodes) == []
    assert p.variables == {"mau": {}, "meta": {"attr": "42"}}


def test_parse_variable_definition_with_multiple_dots():
    p = init_parser(":meta.category.attr:42")
    p.parse()

    assert listasdict(p.nodes) == []
    assert p.variables == {"mau": {}, "meta": {"category.attr": "42"}}


def test_variable_replacement():
    source = """
    :answer:42
    
    The answer is {answer}
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "The answer is 42",
                    }
                ],
            },
        },
    ]

    _test(source, expected)


def test_variable_definition_escape():
    source = r"""
    \:answer:42
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": ":answer:42",
                    }
                ],
            },
        },
    ]

    _test(source, expected)


def test_skip_variable_replacement():
    source = r"""
    :answer:42
    
    The answer is \{answer\}
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "The answer is {answer}",
                    }
                ],
            },
        },
    ]

    _test(source, expected)


def test_variables_are_preprocessed():
    source = """
    :important:*IMPORTANT*
    
    This is {important}
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is ",
                    },
                    {
                        "type": "style",
                        "value": "star",
                        "content": {
                            "type": "sentence",
                            "content": [{"type": "text", "value": "IMPORTANT"}],
                        },
                    },
                ],
            },
        }
    ]

    _test(source, expected)



import copy
import logging

import jinja2

from mau.parsers import nodes
from mau.parsers.main_parser import MainParser

_logger = logging.getLogger(__name__)


class TemplateNotFound(ValueError):
    pass


class CommandError(ValueError):
    pass


def return_node(node, *args, **kwargs):
    return node


class Visitor:
    _template_extension = ".txt"
    _environment_options = {}

    def __init__(
        self,
        default_templates=None,
        custom_templates=None,
        templates_directory=None,
        config=None,
        toc=None,
        footnotes=None,
    ):
        self.default_templates = default_templates or {}
        self.custom_templates = custom_templates or {}
        self.templates_directory = templates_directory
        self.config = copy.deepcopy(config) if config else {}
        self.toc = toc or nodes.TocNode()
        self.footnotes = footnotes or nodes.FootnotesNode()

        self.default_templates.update(self.custom_templates)

        # This is the fallback environment for templates
        self.default_templates_env = jinja2.Environment(
            loader=jinja2.DictLoader(self.default_templates),
            **self._environment_options,
        )

        # This is the first choice for templates
        if self.templates_directory:
            self.main_templates_env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(searchpath=self.templates_directory),
                **self._environment_options,
            )
        else:
            self.main_templates_env = self.default_templates_env

    def _template(self, node_type):  # pragma: no cover
        # This returns a template or raises TemplateNotFound

        template_name = f"{node_type}{self._template_extension}"

        try:
            return self.main_templates_env.get_template(template_name)
        except jinja2.exceptions.TemplateNotFound:
            _logger.debug(
                f"Template {template_name} not provided, trying the default one"
            )
            try:
                return self.default_templates_env.get_template(template_name)
            except jinja2.exceptions.TemplateNotFound as e:
                raise TemplateNotFound(e)

    def _render(self, node_type, **kwargs):
        # This renders a template passing the visitor configuration
        template = self._template(node_type)

        return template.render(config=self.config, **kwargs)

    def _reduce(self, node, attributes, *args, **kwargs):
        for attribute in attributes:
            node[attribute] = self.visit(node[attribute], *args, **kwargs)

    def _reducelist(self, node, attributes, join_with=None, *args, **kwargs):
        for attribute in attributes:
            node[attribute] = self.visitlist(
                node[attribute], join_with, *args, **kwargs
            )

    def visit(self, node, *args, **kwargs):
        # The visitor has to define functions for each node type
        # and those shall return a dictionary of keys.
        #
        # The node types are a list made of the key `node_types` and the node type.
        # This allows a function to return one or more types that have to
        # be used instead of the standard type, which also allows to write
        # generic functions. See _visit_style in HTMLVisitor which works for
        # multiple styles.
        #
        # The rest of the returned keys are passed to the _render function
        # as keys and are thus available in the template.

        if node is None:
            return ""

        method_name = f'_visit_{node["type"]}'

        try:
            method = getattr(self, method_name)
        except AttributeError:
            method = return_node

        node = method(node, *args, **kwargs)

        node_types = []
        node_types.extend(node.pop("node_types", []))
        node_types.append(node["type"])

        for i in node_types:
            try:
                return self._render(i, **node)
            except TemplateNotFound:
                continue

        raise ValueError(f"Cannot find any suitable template among {node_types}")

    def visitlist(self, nodes, join_with=None, *args, **kwargs):
        visited_nodes = [self.visit(i, *args, **kwargs) for i in nodes]

        if join_with is not None:
            return join_with.join(visited_nodes)

        return visited_nodes

    def _visit_macro(self, node):
        node["node_types"] = [f'macro-{node["name"]}']

        return node

    def _visit_block_engine_mau(self, node):
        # If the engine is mau we need to
        # process the content in an isolated parser
        # This can't easily be done in the MainParser
        # class itself because the visitor needs the
        # TOC from the parser.
        p = MainParser().analyse(node["content"])
        ps = MainParser().analyse(node["secondary_content"])

        visitor = self.__class__(
            default_templates=self.default_templates,
            templates_directory=self.templates_directory,
            config=p.variables,
            toc=p.toc,
            footnotes=p.footnotes,
        )
        node["content"] = visitor.visit(nodes.ContainerNode(p.nodes).asdict())
        node["secondary_content"] = visitor.visit(
            nodes.ContainerNode(ps.nodes).asdict()
        )

        return node

    def _visit_block(self, node):
        node["node_types"] = [
            f'block-{node["engine"]}-{node["blocktype"]}',
            f'block-{node["engine"]}',
            f'block-{node["blocktype"]}',
        ]

        self._reducelist(node, ["content"], join_with="\n")
        self._reducelist(node, ["secondary_content"], join_with="\n")
        self._reduce(node, ["title"])

        method = None
        for method_name in [
            f'_visit_block_engine_{node["engine"]}',
            f'_visit_block_{node["blocktype"]}',
        ]:
            try:
                method = getattr(self, method_name)
            except AttributeError:
                continue

        if method:
            return method(node)

        return node

    def _visit_class(self, node):
        self._reduce(node, ["content"])
        return node

    def _visit_container_node(self, node):
        node["node_types"] = ["container"]
        node["content"] = "".join(self.visitlist(node["content"]))
        return node

    def _visit_content_image(self, node):
        node["node_types"] = ["image"]
        self._reduce(node, ["title"])
        return node

    def _visit_document(self, node):
        self._reducelist(node, ["content"], join_with="")
        return node

    def _visit_footnote_def(self, node):
        self._reducelist(node, ["content"], join_with="")
        return node

    def _visit_image(self, node):
        node["node_types"] = ["inline_image"]
        return node

    def _visit_list(self, node):
        self._reducelist(node, ["items"], join_with="")
        return node

    def _visit_list_item(self, node):
        self._reduce(node, ["content"])
        return node

    def _visit_paragraph(self, node):
        self._reduce(node, ["content"])
        return node

    def _visit_quote(self, node):
        self._reduce(node, ["attribution"])
        node["content"] = self.visitlist(node["content"], join_with="")
        return node

    def _visit_raw(self, node):
        node["content"] = self.visitlist(node["content"], join_with="\n")
        return node

    def _visit_sentence(self, node):
        node["content"] = "".join(self.visitlist(node["content"]))
        return node

    def _visit_style(self, node):
        node["node_types"] = [node["value"]]
        self._reduce(node, ["content"])
        return node

    def _visit_toc_entry(self, node):
        self._reducelist(node, ["children"], join_with="")
        return node

    def _visit_verbatim(self, node):
        node["content"] = node["value"]
        return node

    def _visit_command(self, node):
        node["node_types"] = [
            f'command-{node["name"]}',
        ]

        try:
            method = getattr(self, f'_visit_command_{node["name"]}')
        except AttributeError:
            return node

        return method(node)



from copy import deepcopy

from mau.lexers.base_lexer import Token, TokenTypes
from mau.lexers.arguments_lexer import ArgumentsLexer
from mau.parsers.base_parser import BaseParser, ParserError


# The ArgumentsParser is used to parse
# unnamed and named arguments passed to macros.
# It parses a string into a list of unnamed
# arguments and a dictionary of named ones.
class ArgumentsParser(BaseParser):
    def __init__(self):
        super().__init__()
        self.lexer = ArgumentsLexer()

        # This flag is turned on as soon as
        # a named argument is parsed
        self._named_arguments = False

        self.kwargs = {}
        self.args = []

    def set_names_and_defaults(self, positional_names, default_values=None):
        """
        Gives names to positional arguments and assigns
        default values to the ones that have not been
        initialised.
        """

        if default_values:
            _default_values = default_values.copy()
        else:
            _default_values = {}

        # This happens if we passed too many positional
        # values. The case where we pass less positional
        # values than required is checked later when named
        # arguments are merged.
        if len(self.args) > len(positional_names):
            self.error(
                f"Error while parsing arguments {self.args} through names {positional_names}"
            )

        positional_arguments = dict(zip(positional_names, self.args))

        # Named arguments win over the defaults
        _default_values.update(self.kwargs)

        # Positional arguments with win over all the rest
        _default_values.update(positional_arguments)

        # Positional arguments are mandatory and strict
        # so all the names have to be present in the
        # final dictionary.
        if not set(positional_names).issubset(set(_default_values.keys())):
            self.error(
                f"The following attributes need to be specified: {positional_names}"
            )

        self.kwargs = _default_values
        self.args = []

    def pop(self):
        """
        Return the first unnamed argument,
        removing it from the list.
        If no arguments are available it
        returns None.
        """
        try:
            return self.args.pop(0)
        except IndexError:
            return None

    def get_arguments_and_reset(self):
        args = deepcopy(self.args)
        kwargs = deepcopy(self.kwargs)

        self.args = []
        self.kwargs = {}
        self._named_arguments = False

        return args, kwargs

    def _parse_named_argument(self):
        # This parses a named argument
        # in the form name=value or name="value"

        name = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, "=")

        # Values can be surrounded by quotes
        if self.peek_token_is(TokenTypes.LITERAL, '"'):
            self.get_token(TokenTypes.LITERAL, '"')
            value = self.collect_join([Token(TokenTypes.LITERAL, '"')])
            self.get_token(TokenTypes.LITERAL, '"')
        else:
            value = self.collect_join(
                [Token(TokenTypes.LITERAL, ","), Token(TokenTypes.EOF)]
            )

        return name, value

    def _parse_unnamed_argument(self):
        # This parses a named argument
        # in the form value or "value"

        # Values can be surrounded by quotes
        if self.peek_token_is(TokenTypes.LITERAL, '"'):
            self.get_token(TokenTypes.LITERAL, '"')
            value = self.collect_join([Token(TokenTypes.LITERAL, '"')])
            self.get_token(TokenTypes.LITERAL, '"')
        else:
            value = self.collect_join(
                [Token(TokenTypes.LITERAL, ","), Token(TokenTypes.EOF)]
            )

        return value

    def _parse_single_argument(self):
        # This parses a single argument, named or unnamed

        with self:
            name, value = self._parse_named_argument()
            self.kwargs[name] = value
            self._named_arguments = True
            return

        with self:
            value = self._parse_unnamed_argument()

            if self._named_arguments:
                raise ParserError(
                    "Unnamed arguments after named arguments are forbidden"
                )

            self.args.append(value)
            return

    def _parse_arguments(self):
        # Just run until you can't find any more commas
        while True:
            self._parse_single_argument()

            with self:
                self.get_token(TokenTypes.LITERAL, ",")
                with self:
                    self.get_token(TokenTypes.WHITESPACE)
                continue

            break

    def parse(self):
        self._parse_arguments()



from mau.visitors.html_visitor import HTMLVisitor

from mau.parsers import nodes
from mau.parsers.main_parser import MainParser
from tests.helpers import (
    dedent,
    remove_indentation,
    init_parser_factory,
    visitlist_factory,
    ast_test_factory,
)

init_parser = init_parser_factory(MainParser)
visitlist = visitlist_factory(HTMLVisitor)

_test = ast_test_factory(MainParser, HTMLVisitor)


def test_toc_entry():
    v = HTMLVisitor()

    node = nodes.TocEntryNode(
        nodes.HeaderNode("Header 1", 1, "Header 1-XXXXXX"), children=[]
    )

    v._visit_toc_entry(node.asdict())

    assert node.asdict() == {
        "type": "toc_entry",
        "header": {
            "type": "header",
            "value": "Header 1",
            "level": 1,
            "anchor": "Header 1-XXXXXX",
            "kwargs": {},
            "tags": [],
        },
        "children": [],
    }


def test_command_toc():

    toc = nodes.TocNode(
        entries=[
            nodes.TocEntryNode(
                nodes.HeaderNode("Header 1", 1, "Header 1-XXXXXX"),
                children=[
                    nodes.TocEntryNode(
                        nodes.HeaderNode("Header 1.1", 2, "Header 1.1-XXXXXX")
                    )
                ],
            ),
            nodes.TocEntryNode(
                nodes.HeaderNode("Header 2", 1, "Header 2-XXXXXX"),
            ),
        ]
    )

    parser = init_parser("::toc:")
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=toc)

    assert result == [
        remove_indentation(
            """
            <div>
              <ul>
                <li>
                  <a href="#Header 1-XXXXXX">Header 1</a>
                  <ul>
                    <li>
                      <a href="#Header 1.1-XXXXXX">Header 1.1</a>
                    </li>
                  </ul>
                </li>
                <li>
                  <a href="#Header 2-XXXXXX">Header 2</a>
                </li>
              </ul>
            </div>
            """
        )
    ]


def test_command_toc_exclude_tags():
    source = dedent(
        """
        [tags=notoc]
        = Header 1

        == Header 1.1

        = Header 2

        ::toc:exclude_tags=notoc
        """
    )

    parser = init_parser(source)
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=parser.toc)

    assert result == [
        '<h1 id="header-1">Header 1</h1>',
        '<h2 id="header-1.1">Header 1.1</h2>',
        '<h1 id="header-2">Header 2</h1>',
        remove_indentation(
            """
            <div>
              <ul>
                <li>
                  <a href="#header-2">Header 2</a>
                </li>
              </ul>
            </div>
            """
        ),
    ]


def test_command_toc_exclude_tags_multiple_tags():
    toc = nodes.TocNode(
        entries=[
            nodes.TocEntryNode(
                nodes.HeaderNode(
                    "Header 1", 1, "Header 1-XXXXXX", kwargs={"tags": "notoc,section"}
                ),
                children=[
                    nodes.TocEntryNode(
                        nodes.HeaderNode("Header 1.1", 2, "Header 1.1-XXXXXX")
                    )
                ],
            ),
            nodes.TocEntryNode(
                nodes.HeaderNode(
                    "Header 2", 1, "Header 2-XXXXXX", kwargs={"tags": "section"}
                ),
            ),
        ]
    )

    parser = init_parser('::toc:exclude_tags="notoc"')
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=toc)

    assert result == [
        remove_indentation(
            """
            <div>
              <ul>
                <li>
                  <a href="#Header 2-XXXXXX">Header 2</a>
                </li>
              </ul>
            </div>
            """
        )
    ]


def test_toc_generation_without_headers():
    source = dedent(
        """
        ::toc:
        """
    )

    parser = init_parser(source)
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=parser.toc)

    assert result == ["<div></div>"]


def test_toc_generation_from_headers():
    source = dedent(
        """
        = Header 1

        == Header 1.1

        ::toc:
        """
    )

    parser = init_parser(source)
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=parser.toc)

    assert result == [
        '<h1 id="header-1">Header 1</h1>',
        '<h2 id="header-1.1">Header 1.1</h2>',
        remove_indentation(
            """
            <div>
              <ul>
                <li>
                  <a href="#header-1">Header 1</a>
                  <ul>
                    <li>
                      <a href="#header-1.1">Header 1.1</a>
                    </li>
                  </ul>
                </li>
              </ul>
            </div>
            """
        ),
    ]


def test_toc_generation_in_standard_blocks():
    source = dedent(
        """
        = Header 1

        == Header 1.1

        [someblock]
        ----
        == Header 1.2

        ::toc:
        ----
        """
    )

    parser = init_parser(source)
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=parser.toc)

    assert result == [
        '<h1 id="header-1">Header 1</h1>',
        '<h2 id="header-1.1">Header 1.1</h2>',
        '<div class="someblock"><div class="content"><h2 id="header-1.2">Header 1.2</h2>\n<div><ul><li><a href="#header-1">Header 1</a><ul><li><a href="#header-1.1">Header 1.1</a></li><li><a href="#header-1.2">Header 1.2</a></li></ul></li></ul></div></div></div>',
    ]



import copy

from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import get_formatter_by_name

from mau.parsers import nodes
from mau.visitors.visitor import Visitor


DEFAULT_TEMPLATES = {
    "block-admonition.html": (
        '<div class="admonition {{ kwargs.class }}">'
        '<i class="{{ kwargs.icon }}"></i>'
        '<div class="content">'
        '<div class="title">{{ kwargs.label }}</div>'
        "<div>{{ content }}</div>"
        "</div></div>"
    ),
    "block.html": (
        '<div class="{{ blocktype }}{% if classes %} {{ classes }}{% endif %}">'
        '{% if title %}<div class="title">{{ title }}</div>{% endif %}'
        '<div class="content">{{ content }}</div>'
        "</div>"
    ),
    "block-source.html": (
        '<div{% if blocktype %} class="{{ blocktype }}"{% endif %}>'
        '{% if title %}<div class="title">{{ title }}</div>{% endif %}'
        '<div class="content">{{ content }}</div>'
        '{% if kwargs.callouts %}<div class="callouts">'
        "<table><tbody>"
        "{% for callout in kwargs.callouts %}<tr>"
        "<td>{{ callout[0] }}</td>"
        "<td>{{ callout[1] }}</td>"
        "</tr>{% endfor %}"
        "</tbody></table>"
        "</div>{% endif %}"
        "</div>"
    ),
    "callout.html": '<span class="callout">{{ name }}</span>',
    "class.html": '<span class="{{ classes }}">{{ content }}</span>',
    "command.html": "{{ content }}",
    "container.html": "{{ content }}",
    "document.html": "<html><head></head><body>{{ content }}</body></html>",
    "footnote_def.html": (
        '<div id="{{ defanchor }}">'
        '<a href="#{{ refanchor }}">{{ number }}</a> {{ content }}</div>'
    ),
    "footnote_ref.html": (
        "<sup>"
        '[<a id="{{ refanchor }}" href="#{{ defanchor }}">{{ number }}</a>]'
        "</sup>"
    ),
    "footnotes.html": '<div id="_footnotes">{{ entries }}</div>',
    "header.html": '<h{{ level }} id="{{ anchor }}">{{ value }}</h{{ level }}>',
    "horizontal_rule.html": "<hr>",
    "image.html": (
        '<div class="imageblock">'
        '<div class="content">'
        '<img src="{{ uri }}"{% if alt_text %} alt="{{ alt_text }}"{% endif %} />'
        '{% if title %}<div class="title">{{ title }}</div>{% endif %}'
        "</div></div>"
    ),
    "inline_image.html": (
        '<span class="image">'
        '<img src="{{ uri }}"{%if alt_text %} alt="{{ alt_text }}"{% endif %}>'
        "</span>"
    ),
    "link.html": '<a href="{{ target }}">{{ text }}</a>',
    "list.html": (
        "<{% if ordered %}ol{% else %}ul{% endif %}>"
        "{{ items }}"
        "</{% if ordered %}ol{% else %}ul{% endif %}>"
    ),
    "list_item.html": "<li>{{ content }}</li>",
    "macro.html": "",
    "paragraph.html": "<p>{{ content }}</p>",
    "block-quote.html": (
        "<blockquote>"
        "{{ content }}"
        "<cite>{% if kwargs.attribution %}{{ kwargs.attribution }}{% else %}{{ secondary_content }}{% endif %}</cite>"
        "</blockquote>"
    ),
    "raw.html": "{{ content }}",
    "sentence.html": "{{ content }}",
    "star.html": "<strong>{{ content }}</strong>",
    "text.html": "{{ value }}",
    "toc.html": "<div>{% if entries%}<ul>{{ entries }}</ul>{% endif %}</div>",
    "toc_entry.html": (
        "<li>"
        '<a href="#{{ header.anchor }}">{{ header.value }}</a>'
        "{% if children %}<ul>{{ children }}</ul>{% endif %}"
        "</li>"
    ),
    "underscore.html": "<em>{{ content }}</em>",
    "verbatim.html": "<code>{{ content }}</code>",
}


class HTMLVisitor(Visitor):
    _template_extension = ".html"

    def __init__(
        self,
        default_templates=None,
        custom_templates=None,
        templates_directory=None,
        config=None,
        toc=None,
        footnotes=None,
    ):
        default_templates = (
            copy.deepcopy(default_templates)
            if default_templates is not None
            else copy.deepcopy(DEFAULT_TEMPLATES)
        )

        super().__init__(
            default_templates=default_templates,
            custom_templates=custom_templates,
            templates_directory=templates_directory,
            config=config,
            toc=toc,
            footnotes=footnotes,
        )

    def _visit_class(self, node):
        node["classes"] = " ".join(node["classes"])
        return super()._visit_class(node)

    def _visit_admonition(self, node):
        node["node_types"] = [f'admonition_{node["class"]}']
        self._reducelist(node, ["content"], join_with="")
        return node

    def _visit_block_engine_source(self, node):
        # The Pygments lexer for the given language
        lexer = get_lexer_by_name(node["kwargs"]["language"])

        # Fetch global configuration for Pygments and for the HtmlFormatter
        pygments_config = self.config.get("pygments", {})
        formatter_config = pygments_config.get("html", {})

        # Get all the attributes of this specific block
        # that start with `pygments.`
        node_pygments_config = dict(
            (k.replace("pygments.", ""), v)
            for k, v in node["kwargs"].items()
            if k.startswith("pygments.")
        )

        # Converting from text to Python might be tricky,
        # so for now I just update the formatter config with
        # 'hl_lines' which is a list of comma-separated integers
        hl_lines = node_pygments_config.get("hl_lines", "")
        hl_lines = hl_lines.split(",")

        # Add all lines that have been highlighted with Mau's custom callouts
        hl_lines = [i for i in hl_lines if i != ""]

        # Pygments starts counting form 1, Mau from 0
        highlight_markers = [i + 1 for i in node["kwargs"]["highlights"]]

        # Merge the two
        hl_lines = list(set(hl_lines) | set(highlight_markers))

        # Tell Pygments which lines we want to highlight
        formatter_config["hl_lines"] = hl_lines

        # Create the formatter and pass the config
        formatter = get_formatter_by_name("html", **formatter_config)

        # Get the raw source from the node
        src = node["content"]

        # Highlight the source with Pygments
        highlighted_src = highlight(src, lexer, formatter)

        # Split it again into lines to work out callouts
        highlighted_src_lines = highlighted_src.split("\n")

        # Inject callout markers in the highlighted code
        callout_markers = node["kwargs"]["callouts"]["markers"]
        callout_contents = node["kwargs"]["callouts"]["contents"]

        for linenum, callout_name in callout_markers.items():
            highlighted_src_lines[linenum] = "{line} {callout}".format(
                line=highlighted_src_lines[linenum],
                callout=self._render("callout", name=callout_name),
            )

        # Rebuild the source code part and the list of callouts
        highlighted_src = "\n".join(highlighted_src_lines)
        callouts_list = [
            (self._render("callout", name=name), text)
            for name, text in callout_contents.items()
        ]

        node["content"] = highlighted_src
        node["kwargs"]["callouts"] = callouts_list

        return node

    def _visit_toc_entry(self, node, exclude_tags=None):
        exclude_tags = exclude_tags or []

        tags_excluded = [i for i in node["header"]["tags"] if i in exclude_tags]

        if len(tags_excluded) > 0:
            # Return an empty text node
            return nodes.TextNode(value="").asdict()

        self._reducelist(node, ["children"], join_with="", exclude_tags=exclude_tags)
        return node

    def _visit_command_toc(self, node):
        exclude_tags = node["kwargs"].get("exclude_tags", [])
        toc_node = self.toc.asdict()

        self._reducelist(toc_node, ["entries"], join_with="", exclude_tags=exclude_tags)

        return toc_node

    def _visit_command_footnotes(self, node):
        footnotes_node = self.footnotes.asdict()
        self._reducelist(footnotes_node, ["entries"], join_with="")

        return footnotes_node



import pytest

from mau.parsers.base_parser import ParserError
from mau.parsers.arguments_parser import ArgumentsParser

from tests.helpers import init_parser_factory

init_parser = init_parser_factory(ArgumentsParser)


def test_single_unnamed_argument_no_spaces():
    p = init_parser("value")
    p.parse()

    assert p.args == ["value"]
    assert p.kwargs == {}


def test_single_unnamed_argument_with_spaces():
    p = init_parser("value with spaces")
    p.parse()

    assert p.args == ["value with spaces"]
    assert p.kwargs == {}


def test_multiple_unnamed_arguments_no_spaces():
    p = init_parser("value1,value2")
    p.parse()

    assert p.args == ["value1", "value2"]
    assert p.kwargs == {}


def test_multiple_unnamed_arguments_with_spaces():
    p = init_parser("value1 with spaces,value2 with other spaces")
    p.parse()

    assert p.args == ["value1 with spaces", "value2 with other spaces"]
    assert p.kwargs == {}


def test_multiple_unnamed_arguments_space_after_comma_is_removed():
    p = init_parser("value1 with spaces, value2 with other spaces")
    p.parse()

    assert p.args == ["value1 with spaces", "value2 with other spaces"]
    assert p.kwargs == {}


def test_single_unnamed_argument_with_quotes_no_spaces():
    p = init_parser('"value"')
    p.parse()

    assert p.args == ["value"]
    assert p.kwargs == {}


def test_single_unnamed_argument_with_quotes_with_spaces():
    p = init_parser('"value with spaces"')
    p.parse()

    assert p.args == ["value with spaces"]
    assert p.kwargs == {}


def test_single_unnamed_argument_comma_is_ignored_between_quotes():
    p = init_parser('"value1,value2"')
    p.parse()

    assert p.args == ["value1,value2"]
    assert p.kwargs == {}


def test_multiple_unnamed_arguments_with_quotes_no_spaces():
    p = init_parser('"value1","value2"')
    p.parse()

    assert p.args == ["value1", "value2"]
    assert p.kwargs == {}


def test_multiple_unnamed_arguments_with_quotes_with_spaces():
    p = init_parser('"value1 with spaces","value2 with other spaces"')
    p.parse()

    assert p.args == ["value1 with spaces", "value2 with other spaces"]
    assert p.kwargs == {}


def test_single_unnamed_argument_with_non_delimiting_quotes():
    p = init_parser(r'value "with quotes"')
    p.parse()

    assert p.args == ['value "with quotes"']
    assert p.kwargs == {}


def test_single_unnamed_argument_with_escaped_quotes():
    p = init_parser(r'"value \"with escaped quotes\""')
    p.parse()

    assert p.args == ['value "with escaped quotes"']
    assert p.kwargs == {}


def test_single_named_argument():
    p = init_parser("name=value")
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name": "value"}


def test_single_named_argument_with_spaces():
    p = init_parser("name=value with spaces")
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name": "value with spaces"}


def test_multiple_named_arguments():
    p = init_parser("name1=value1,name2=value2")
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name1": "value1", "name2": "value2"}


def test_multiple_named_arguments_with_spaces():
    p = init_parser("name1=value1 with spaces,name2=value2 with spaces")
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name1": "value1 with spaces", "name2": "value2 with spaces"}


def test_multiple_named_arguments_space_after_comma_is_removed():
    p = init_parser("name1=value1, name2=value2")
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name1": "value1", "name2": "value2"}


def test_single_named_argument_with_quotes_no_spaces():
    p = init_parser('name="value"')
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name": "value"}


def test_single_named_argument_with_quotes_with_spaces():
    p = init_parser('name="value with spaces"')
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name": "value with spaces"}


def test_single_named_argument_comma_is_ignored_between_quotes():
    p = init_parser('name="value1,value2"')
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name": "value1,value2"}


def test_multiple_named_arguments_with_quotes_no_spaces():
    p = init_parser('name1="value1",name2="value2"')
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name1": "value1", "name2": "value2"}


def test_multiple_named_arguments_with_quotes_with_spaces():
    p = init_parser('name1="value1 with spaces",name2="value2 with other spaces"')
    p.parse()

    assert p.args == []
    assert p.kwargs == {
        "name1": "value1 with spaces",
        "name2": "value2 with other spaces",
    }


def test_single_named_argument_with_non_delimiting_quotes():
    p = init_parser(r'name=value "with quotes"')
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name": 'value "with quotes"'}


def test_single_named_argument_with_escaped_quotes():
    p = init_parser(r'name="value \"with escaped quotes\""')
    p.parse()

    assert p.args == []
    assert p.kwargs == {"name": 'value "with escaped quotes"'}


def test_apply_prototype_unnamed_arguments():
    p = init_parser("value1, value2")
    p.parse()
    p.set_names_and_defaults(["attr1", "attr2"], {"attr3": "value3"})

    assert p.args == []
    assert p.kwargs == {"attr1": "value1", "attr2": "value2", "attr3": "value3"}


def test_apply_prototype_unnamed_and_named_arguments():
    p = init_parser("value1, value2, attr5=value5")
    p.parse()
    p.set_names_and_defaults(["attr1", "attr2"], {"attr3": "value3"})

    assert p.args == []
    assert p.kwargs == {
        "attr1": "value1",
        "attr2": "value2",
        "attr3": "value3",
        "attr5": "value5",
    }


def test_apply_prototype_clash_between_default_and_named():
    p = init_parser("value1, value2, attr3=value5")
    p.parse()
    p.set_names_and_defaults(["attr1", "attr2"], {"attr3": "value3"})

    assert p.args == []
    assert p.kwargs == {
        "attr1": "value1",
        "attr2": "value2",
        "attr3": "value5",
    }


def test_apply_prototype_clash_between_positional_and_named():
    p = init_parser("value1, value2, attr1=value5")
    p.parse()
    p.set_names_and_defaults(["attr1", "attr2"], {"attr3": "value3"})

    assert p.args == []
    assert p.kwargs == {
        "attr1": "value1",
        "attr2": "value2",
        "attr3": "value3",
    }


def test_apply_prototype_clash_in_prototype():
    p = init_parser("value1, value2, attr3=value3")
    p.parse()
    p.set_names_and_defaults(["attr1", "attr2"], {"attr1": "value7"})

    assert p.args == []
    assert p.kwargs == {
        "attr1": "value1",
        "attr2": "value2",
        "attr3": "value3",
    }


def test_apply_prototype_positional_values_not_provided():
    p = init_parser("value1")
    p.parse()

    with pytest.raises(ParserError):
        p.set_names_and_defaults(["attr1", "attr2"], {"attr3": "value3"})


def test_apply_prototype_too_many_positional_values():
    p = init_parser("value1, value2")
    p.parse()

    with pytest.raises(ParserError):
        p.set_names_and_defaults(["attr1"], {"attr3": "value3"})


def test_apply_prototype_missing_positional_with_default():
    p = init_parser("attr1=value5")
    p.parse()
    p.set_names_and_defaults(["attr1"], {})

    assert p.args == []
    assert p.kwargs == {
        "attr1": "value5",
    }



import pytest

from tests.helpers import dedent

from mau import text_buffer
from mau.lexers.base_lexer import BaseLexer, LexerError, EOL, EOF, Text, WS


def test_no_text():
    lex = BaseLexer()

    lex.process("")

    assert lex.tokens == [EOL, EOF]


def test_empty_lines():
    lex = BaseLexer()

    lex.process("\n")

    assert lex.tokens == [EOL, EOL, EOF]
    assert [t.position for t in lex.tokens] == [
        (0, 0),
        (1, 0),
        (2, 0),
    ]


def test_initial_position():
    lex = BaseLexer(initial_position=(1, 5))

    lex.process("\n")

    assert lex.tokens == [EOL, EOL, EOF]
    assert [t.position for t in lex.tokens] == [
        (1, 5),
        (2, 0),
        (3, 0),
    ]


def test_process_letters():
    lex = BaseLexer()

    lex.process("abcd")

    assert lex.tokens == [Text("a"), Text("b"), Text("c"), Text("d"), EOL, EOF]
    assert [t.position for t in lex.tokens] == [
        (0, 0),
        (0, 1),
        (0, 2),
        (0, 3),
        (0, 4),
        (1, 0),
    ]


def test_process_whitespace():
    lex = BaseLexer()

    lex.process("a   b")

    assert lex.tokens == [Text("a"), WS("   "), Text("b"), EOL, EOF]


def test_lexer_error():
    lex = BaseLexer()

    with pytest.raises(LexerError):
        lex.process("123")


####################################################
# Private methods
#
# These normally shouldn't be tested, but this class
# is the foundation of all other lexers, so I prefer
# to keep an eye on the internal methods, as they
# will be used by the child lexers.


def test_wrap():
    lex = BaseLexer()

    assert lex._wrap(None) is None
    assert lex._wrap([5, 6]) == [5, 6]
    assert lex._wrap(5) == [5]


def test_current_char_and_line():
    lex = BaseLexer()
    buf = text_buffer.TextBuffer()
    buf.load("Just an example\nOn two lines")
    lex._text_buffer = buf

    assert lex._current_char == "J"
    assert lex._current_line == "Just an example"


def test_insert():
    lex = BaseLexer()
    buf = text_buffer.TextBuffer()
    buf.load("Just an example\nOn two lines")
    lex._text_buffer = buf
    lex._nextline()

    lex._insert("A third line!")

    assert lex._current_line == "A third line!"
    lex._nextline()
    assert lex._current_line == "On two lines"



import re
import copy

from mau.lexers.base_lexer import TokenTypes, Token
from mau.lexers.main_lexer import MainLexer
from mau.parsers.base_parser import (
    BaseParser,
    TokenError,
    ConfigurationError,
    parser,
)
from mau.parsers.text_parser import TextParser
from mau.parsers.arguments_parser import ArgumentsParser
from mau.parsers.preprocess_variables_parser import PreprocessVariablesParser
from mau.parsers.nodes import (
    HorizontalRuleNode,
    TextNode,
    BlockNode,
    ContentNode,
    ContentImageNode,
    CommandNode,
    HeaderNode,
    ListNode,
    ListItemNode,
    ParagraphNode,
    TocNode,
    TocEntryNode,
    FootnotesNode,
)


class EngineError(ValueError):
    """ Used to signal that the engine selected for a code block is not known """


def header_anchor(text, level):
    """
    Return a sanitised anchor for a header.
    """

    # Everything lowercase
    sanitised_text = text.lower()

    # Get only letters, numbers, dashes, spaces, and dots
    sanitised_text = "".join(re.findall("[a-z0-9-\\. ]+", sanitised_text))

    # Remove multiple spaces
    sanitised_text = "-".join(sanitised_text.split())

    return sanitised_text


# The MainParser is in charge of parsing
# the whole input, calling other parsers
# to manage single paragraphs or other
# things like variables.
class MainParser(BaseParser):
    def __init__(self, variables=None):
        super().__init__()

        self.lexer = MainLexer()

        # This is used as a storage for attributes.
        # Block attributes are defined before the block
        # so when we parse them we store them here and
        # then use them when dealing with the block itself.
        self.argsparser = ArgumentsParser()

        # Copy the variables and make sure the "mau" namespace exists
        self.variables = copy.deepcopy(variables) if variables else {}
        if "mau" not in self.variables:
            self.variables["mau"] = {}

        self.headers = []
        self.footnote_defs = []
        self.blocks = {}
        self.toc = None

        # When we define a block we establish an alias
        # {alias:actual_block_name}
        self.block_aliases = {}

        # Each block we define can have default values
        # {actual_block_name:kwargs}
        self.block_defaults = {}

        # Each block we define can have names for unnamed arguments
        # {actual_block_name:kwargs}
        self.block_names = {}

        # Backward compatibility with Mau 1.x
        # Mau 1.x used [source] to format source, while Mau 2.x
        # uses [myblock, engine=source], so this establishes
        # a default block definition so that
        # [source] = [source, engine=source]
        # In Mau 2.x this block uses the template "block-source"
        # so any template called "source" (e.g. "source.html")
        # must be renamed.
        # This definition can be overridden by custom block definitions
        self.block_aliases["source"] = "source"
        self.block_defaults["source"] = {"engine": "source", "language": "text"}
        self.block_names["source"] = ["language"]

        self.block_aliases["admonition"] = "admonition"
        self.block_names["admonition"] = ["class", "icon", "label"]

        self.block_aliases["quote"] = "quote"
        self.block_defaults["quote"] = {"attribution": None}
        self.block_names["quote"] = ["attribution"]

        # Iterate through block definitions passed as variables
        for alias, block_definition in (
            self.variables["mau"].get("block_definitions", {}).items()
        ):
            try:
                blocktype = block_definition["blocktype"]
                self.block_aliases[alias] = blocktype
            except KeyError:
                raise ConfigurationError(
                    f"Block definition '{alias}' is missing key 'blocktype'"
                )

            try:
                self.block_defaults[blocktype] = block_definition["kwargs"]
            except KeyError:
                raise ConfigurationError(
                    f"Block definition '{alias}' is missing key 'kwargs'"
                )

        # This is a buffer for a block title
        self._title = None

        # This is the function used to create the header
        # anchors. It can be specified through
        # mau.header_anchor_function to override
        # the default one.
        self.header_anchor = self.variables["mau"].get(
            "header_anchor_function", header_anchor
        )

        self.v1_backward_compatibility = self.variables["mau"].get(
            "v1_backward_compatibility", False
        )

    def _pop_title(self):
        # This return the title and resets the
        # cached one, so no other block will
        # use it.
        title = self._title
        self._title = None
        return title

    def _push_title(self, title):
        # When we parse a title we can store it here
        # so that it is available to the next block
        # that will use it.
        self._title = title

    def _collect_lines(self, stop_tokens):
        # This collects several lines of text in a list
        # until it gets to a line that begins with one
        # of the tokens listed in stop_tokens.
        # It is useful for block or other elements that
        # are clearly surrounded by delimiters.
        lines = []

        while self.peek_token() not in stop_tokens:
            lines.append(self.collect_join([Token(TokenTypes.EOL)]))
            self.get_token(TokenTypes.EOL)

        return lines

    def _collect_text_content(self):
        # Collects all adjacent text tokens
        # into a single string

        if not self.peek_token_is(TokenTypes.TEXT):
            return None

        values = []

        # Get all tokens
        while self.peek_token_is(TokenTypes.TEXT):
            values.append(self.get_token().value)
            self.get_token(TokenTypes.EOL)

        return " ".join(values)

    def _parse_text_content(self, text):
        # Parse a text using the TextParser.

        # Replace variables
        p = PreprocessVariablesParser(self.variables).analyse(
            text,
        )
        text = p.nodes[0].value

        # Parse the text
        p = TextParser(
            footnotes_start_with=len(self.footnote_defs) + 1,
            v1_backward_compatibility=self.v1_backward_compatibility,
        ).analyse(text)

        # Text should return a single sentence node
        result = p.nodes[0]

        # Store the footnotes
        self.footnote_defs.extend(p.footnote_defs)

        return result

    @parser
    def _parse_eol(self):
        # This simply parses the end of line.

        self.get_token(TokenTypes.EOL)

    @parser
    def _parse_horizontal_rule(self):
        # The horizontal rule ---

        self.get_token(TokenTypes.LITERAL, "---")
        self.get_token(TokenTypes.EOL)

        self._save(HorizontalRuleNode())

    @parser
    def _parse_single_line_comment(self):
        # // A comment on a single line

        self.get_token(TokenTypes.TEXT, check=lambda x: x.startswith("//"))
        self.get_token(TokenTypes.EOL)

    @parser
    def _parse_multi_line_comment(self):
        # ////
        # A comment
        # on multiple lines
        # ////

        self.get_token(TokenTypes.LITERAL, "////")
        self._collect_lines([Token(TokenTypes.LITERAL, "////"), Token(TokenTypes.EOF)])
        self.force_token(TokenTypes.LITERAL, "////")

    @parser
    def _parse_variable_definition(self):
        # This parses a variable definition
        #
        # Simple variables are defined as :name:value
        # as True booleans as just :name:
        # and as False booleas as :!name:
        #
        # Variable names can use a namespace with
        # :namespace.name:value

        # Get the mandatory variable name
        self.get_token(TokenTypes.LITERAL, ":")
        variable_name = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, ":")

        # Assume the variable is a flag
        variable_value = True

        # If the name starts with ! it's a false flag
        if variable_name.startswith("!"):
            variable_value = False
            variable_name = variable_name[1:]

        # Get the optional value
        value = self.collect_join([Token(TokenTypes.EOL)])

        # The value is assigned only if the variable
        # is not a negative flag. In that case it is ignored
        if variable_value and len(value) > 0:
            variable_value = value

        # If the variable name contains a dot we
        # want to use a namespace
        if "." not in variable_name:
            self.variables[variable_name] = variable_value
        else:
            # Let's ignore all others dots
            namespace, variable_name = variable_name.split(".", maxsplit=1)

            # This defines the namespace if it's not already there
            try:
                self.variables[namespace][variable_name] = variable_value
            except KeyError:
                self.variables[namespace] = {variable_name: variable_value}

    @parser
    def _parse_command(self):
        # Parse a command in the form ::command:

        self.get_token(TokenTypes.LITERAL, "::")
        name = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, ":")

        args = []
        kwargs = {}

        # Commands can have arguments
        with self:
            arguments = self.get_token(TokenTypes.TEXT).value
            self.argsparser.analyse(arguments)

            # Consume the attributes
            args, kwargs = self.argsparser.get_arguments_and_reset()

        if name == "defblock":
            # Block definitions must have at least 2 arguments,
            # the alias and the block type.
            if len(args) < 2:
                self.error(
                    "Block definitions require at least two unnamed arguments: ALIAS and BLOCKTYPE"
                )

            block_alias = args.pop(0)
            block_type = args.pop(0)

            self.block_aliases[block_alias] = block_type
            self.block_defaults[block_type] = kwargs
            self.block_names[block_type] = args

            return None

        self._save(CommandNode(name=name, args=args, kwargs=kwargs))

    @parser
    def _parse_title(self):
        # Parse a title in the form
        #
        # . This is a title
        # or
        # .This is a title

        # Parse the mandatory dot
        self.get_token(TokenTypes.LITERAL, ".")

        # Parse the optional white spaces
        with self:
            self.get_token(TokenTypes.WHITESPACE)

        # Get the text of the title
        text = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.EOL)

        # Titles can contain Mau code
        p = TextParser(
            footnotes_start_with=len(self.footnote_defs) + 1,
            v1_backward_compatibility=self.v1_backward_compatibility,
        ).analyse(text)
        title = p.nodes[0]

        self._push_title(title)

    @parser
    def _parse_attributes(self):
        # Parse block attributes in the form
        # [unnamed1, unnamed2, ..., named1=value1, name2=value2, ...]

        self.get_token(TokenTypes.LITERAL, "[")
        attributes = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, "]")

        # Attributes can use variables
        p = PreprocessVariablesParser(self.variables).analyse(
            attributes,
        )
        attributes = p.nodes[0].value

        # Parse the arguments
        self.argsparser.analyse(attributes)

    @parser
    def _parse_header(self):
        # Parse a header in the form
        #
        # = Header
        #
        # The number of equal signs is arbitrary
        # and represents the level of the header.
        # Headers are automatically assigned an anchor
        # created using the provided function self.header_anchor
        #
        # Headers in the form
        # =! Header
        # are rendered but not included in the TOC

        # Get all the equal signs
        header = self.get_token(
            TokenTypes.LITERAL, check=lambda x: x.startswith("=")
        ).value

        # Get the mandatory white spaces
        self.get_token(TokenTypes.WHITESPACE)

        # Check if the header has to be in the TOC
        in_toc = True
        if header.endswith("!"):
            header = header[:-1]
            in_toc = False

        # Get the text of the header and calculate the level
        text = self.get_token(TokenTypes.TEXT).value
        level = len(header)

        # Generate the anchor and append it to the TOC
        anchor = self.header_anchor(text, level)

        # Consume the attributes
        args, kwargs = self.argsparser.get_arguments_and_reset()

        # Generate the header node
        header_node = HeaderNode(value=text, level=level, anchor=anchor, kwargs=kwargs)

        if in_toc:
            self.headers.append(header_node)

        self._save(header_node)

    @parser
    def _parse_block(self):
        # Parse a block in the form
        #
        # [block_type]
        # ----
        # Content
        # ----
        # Optional secondary content
        #
        # Blocks are delimited by 4 consecutive identical characters.

        # Get the delimiter and check the length
        delimiter = self.get_token(TokenTypes.TEXT).value
        if len(delimiter) != 4 or len(set(delimiter)) != 1:
            raise TokenError
        self.get_token(TokenTypes.EOL)

        # Collect everything until the next delimiter
        content = self._collect_lines(
            [Token(TokenTypes.TEXT, delimiter), Token(TokenTypes.EOF)]
        )
        self.force_token(TokenTypes.TEXT, delimiter)
        self.get_token(TokenTypes.EOL)

        # Get the optional secondary content
        secondary_content = self._collect_lines(
            [Token(TokenTypes.EOL), Token(TokenTypes.EOF)]
        )

        # Consume the title
        title = self._pop_title()

        # The first unnamed argument is the block type
        blocktype = self.argsparser.pop()

        # If there is a block alias for blocktype replace it
        # otherwise use the blocktype we already have
        blocktype = self.block_aliases.get(blocktype, blocktype)

        # Assign names

        self.argsparser.set_names_and_defaults(
            self.block_names.get(blocktype, []), self.block_defaults.get(blocktype, {})
        )

        # Consume the attributes
        args, kwargs = self.argsparser.get_arguments_and_reset()

        # Extract classes and convert them into a list
        classes = [i for i in kwargs.pop("classes", "").split(",") if len(i) > 0]

        # Extract condition if present and process it
        condition = kwargs.pop("condition", "")

        # Run this only if there is a condition on this block
        if len(condition) > 0:
            try:
                # The condition should be either test:variable:value or test:variable:
                test, variable, value = condition.split(":")
            except ValueError:
                self.error(
                    f'Condition {condition} is not in the form "test:variable:value" or "test:variable:'
                )

            # If there is no value use True
            if len(value) == 0:
                value = True

            # Check if the variable matches the value and apply the requested test
            match = self.variables.get(variable) == value
            result = True if test == "if" else False

            # If the condition is not satisfied return
            if match is not result:
                return

        # Extract the preprocessor
        preprocessor = kwargs.pop("preprocessor", "none")

        # Extract the engine
        engine = kwargs.pop("engine", "default")

        # Create the node parameters according to the engine
        if engine in ["raw", "mau"]:
            # Engine "raw" doesn't process the content,
            # so we just pass it untouched in the form of
            # a TextNode per line. The same is true for "mau"
            # as the visitor will have to fire up an new parser
            # to process the content.
            content = [TextNode(line) for line in content]
            secondary_content = [TextNode(line) for line in secondary_content]
        elif engine == "source":
            # Engine "source" extracts the content (source code),
            # the callouts, and the highlights.
            # The default language is "text".

            content, callouts, highlights = self._parse_source_engine(
                content, secondary_content, kwargs
            )
            secondary_content = []

            kwargs["callouts"] = callouts
            kwargs["highlights"] = highlights
            kwargs["language"] = kwargs.get("language", "text")

        elif engine == "default":
            # This is the default engine and it parses
            # both content and secondary content using a new parser
            # but then merges headers and footnotes into the
            # current one.

            # Parse the primary and secondary content and record footnotes
            pc = MainParser(variables=self.variables).analyse("\n".join(content))
            ps = MainParser(variables=self.variables).analyse(
                "\n".join(secondary_content)
            )
            content = pc.nodes
            secondary_content = ps.nodes

            self.footnote_defs.extend(pc.footnote_defs)
            self.headers.extend(pc.headers)
        else:
            raise EngineError(f"Engine {engine} is not available")

        self._save(
            BlockNode(
                blocktype=blocktype,
                content=content,
                secondary_content=secondary_content,
                args=args,
                classes=classes,
                engine=engine,
                preprocessor=preprocessor,
                kwargs=kwargs,
                title=title,
            )
        )

    def _parse_source_engine(self, content, secondary_content, kwargs):
        # Parse a source block in the form
        #
        # [source, language, attributes...]
        # ----
        # content
        # ----
        #
        # Source blocks support the following attributes
        #
        # callouts=":" The separator used by callouts
        # highlight="@" The special character to turn on highlight
        #
        # [source, language, attributes...]
        # ----
        # content:1:
        # ----
        #
        # [source, language, attributes...]
        # ----
        # content:@:
        # ----
        #
        # Callout descriptions can be added to the block
        # as secondary content with the syntax
        #
        # [source, language, attributes...]
        # ----
        # content:name:
        # ----
        # <name>: <description>
        #
        # Since Mau uses Pygments, the attribute language
        # is one of the langauges supported by that tool.

        # Get the delimiter for callouts (":" by default)
        delimiter = kwargs.pop("callouts", ":")

        # A dictionary that contains callout markers in
        # the form {linenum:name}
        callout_markers = {}

        # Get the marker for highlighted lines ("@" by default)
        highlight_marker = kwargs.pop("highlight", "@")

        # A list of highlighted lines
        highlighted_lines = []

        # This is a list of all lines that might contain
        # a callout. They will be further processed
        # later to be sure.
        lines_with_callouts = [
            (linenum, line)
            for linenum, line in enumerate(content)
            if line.endswith(delimiter)
        ]

        # Each line in the previous list is processed
        # and stored if it contains a callout
        for linenum, line in lines_with_callouts:
            # Remove the final delimiter
            line = line[:-1]

            splits = line.split(delimiter)
            if len(splits) < 2:
                # It's a trap! There are no separators left
                continue

            # Get the callout and the line
            callout_name = splits[-1]
            line = delimiter.join(splits[:-1])

            content[linenum] = line

            # Check if we want to just highlight the line
            if callout_name == highlight_marker:
                highlighted_lines.append(linenum)
            else:
                callout_markers[linenum] = callout_name

        # A dictionary that contains the text for each
        # marker in the form {name:text}
        callout_contents = {}

        # If there was secondary content it should be formatted
        # with callout names followed by colon and the
        # callout text.
        for line in secondary_content:
            if ":" not in line:
                self.error(
                    f"Callout description should be written as 'name: text'. Missing ':' in '{line}'"
                )

            name, text = line.split(":")

            if name not in callout_markers.values():
                self.error(f"Callout {name} has not been created in the source code")

            text = text.strip()

            callout_contents[name] = text

        # Put markers and contents together
        callouts = {"markers": callout_markers, "contents": callout_contents}

        # Source blocks must preserve the content literally
        textlines = [TextNode(line) for line in content]

        return textlines, callouts, highlighted_lines

        # self._save(
        #     SourceNode(
        #         language,
        #         callouts=callouts,
        #         highlights=highlighted_lines,
        #         delimiter=delimiter,
        #         code=textlines,
        #         title=title,
        #         kwargs=kwargs,
        #     )
        # )

    @parser
    def _parse_content(self):
        # Parse attached content in the form
        #
        # [attributes]
        # << content_type:uri

        # Get the mandatory "<<" and white spaces
        self.get_token(TokenTypes.LITERAL, check=lambda x: x.startswith("<<"))
        self.get_token(TokenTypes.WHITESPACE)

        # Get the content type and the content URI
        content_type_and_uri = self.get_token(TokenTypes.TEXT).value
        content_type, uri = content_type_and_uri.split(":", maxsplit=1)

        title = self._pop_title()

        if content_type == "image":
            return self._parse_content_image(uri, title)

        return self._parse_standard_content(content_type, uri, title)

    def _parse_content_image(self, uri, title):
        # Parse a content image in the form
        #
        # [alt_text, classes]
        # << image:uri
        #
        # alt_text is the alternate text to use is the image is not reachable
        # and classes is a comma-separated list of classes

        # Assign names and consume the attributes
        self.argsparser.set_names_and_defaults(
            ["alt_text", "classes"], {"alt_text": None, "classes": None}
        )
        args, kwargs = self.argsparser.get_arguments_and_reset()

        alt_text = kwargs.pop("alt_text")
        classes = kwargs.pop("classes")

        if classes:
            classes = classes.split(",")

        self._save(
            ContentImageNode(
                uri=uri,
                alt_text=alt_text,
                classes=classes,
                title=title,
                kwargs=kwargs,
            )
        )

    def _parse_standard_content(self, content_type, uri, title):
        # This is the fallback for an unknown content type

        # Consume the attributes
        args, kwargs = self.argsparser.get_arguments_and_reset()

        self._save(
            ContentNode(
                uri=uri,
                title=title,
                args=args,
                kwargs=kwargs,
            )
        )

    @parser
    def _parse_list(self):
        # Parse a list.
        # Lists can be ordered (using numbers)
        #
        # * One item
        # * Another item
        #
        # or unordered (using bullets)
        #
        # # Item 1
        # # Item 2
        #
        # The number of headers increases
        # the depth of each item
        #
        # # Item 1
        # ## Sub-Item 1.1
        #
        # Spaces before and after the header are ignored.
        # So the previous list can be also written
        #
        # # Item 1
        #   ## Sub-Item 1.1
        #
        # Ordered and unordered lists can be mixed.
        #
        # * One item
        # ## Sub Item 1
        # ## Sub Item 2
        #

        # Ignore initial white spaces
        with self:
            self.get_token(TokenTypes.WHITESPACE)

        # Get the header and decide if it's a numbered or unnumbered list
        header = self.peek_token(TokenTypes.LITERAL, check=lambda x: x[0] in "*#")
        numbered = True if header.value[0] == "#" else False

        # Parse all the following items
        nodes = self._parse_list_nodes()

        self._save(ListNode(numbered, nodes, main_node=True))

    def _parse_list_nodes(self):
        # This parses all items of a list

        # Ignore initial white spaces
        with self:
            self.get_token(TokenTypes.WHITESPACE)

        # Parse the header and ignore the following white spaces
        header = self.get_token(TokenTypes.LITERAL, check=lambda x: x[0] in "*#").value
        self.get_token(TokenTypes.WHITESPACE)

        # Collect and parse the text of the item
        text = self._collect_text_content()
        content = self._parse_text_content(text)

        # Compute the level of the item
        level = len(header)

        nodes = []
        nodes.append(ListItemNode(level, content))

        while not self.peek_token() in [Token(TokenTypes.EOF), Token(TokenTypes.EOL)]:
            # This is the SentenceNode inside the last node added to the list
            # which is used to append potential nested nodes
            last_node_sentence = nodes[-1].content

            # Ignore the initial white spaces
            with self:
                self.get_token(TokenTypes.WHITESPACE)

            if len(self.peek_token().value) == level:
                # The new item is on the same level

                # Get the header
                header = self.get_token().value

                # Ignore white spaces
                self.get_token(TokenTypes.WHITESPACE)

                # Collect and parse the text of the item
                text = self._collect_text_content()
                content = self._parse_text_content(text)

                nodes.append(ListItemNode(len(header), content))
            elif len(self.peek_token().value) > level:
                # The new item is on a deeper level

                # Treat the new line as a new list
                numbered = True if self.peek_token().value[0] == "#" else False
                subnodes = self._parse_list_nodes()
                last_node_sentence.content.append(ListNode(numbered, subnodes))
            else:
                break

        return nodes

    @parser
    def _parse_paragraph(self):
        # This parses a paragraph.
        # Paragraphs can be written on multiple lines and
        # end with an empty line.

        # Get all the lines, join them and parse them
        lines = self._collect_lines([Token(TokenTypes.EOL), Token(TokenTypes.EOF)])
        text = " ".join(lines)
        sentence = self._parse_text_content(text)

        # Consume the attributes
        args, kwargs = self.argsparser.get_arguments_and_reset()

        self._save(ParagraphNode(sentence, args=args, kwargs=kwargs))

    def _parse_functions(self):
        # All the functions that this parser provides.

        return [
            self._parse_eol,
            self._parse_horizontal_rule,
            self._parse_single_line_comment,
            self._parse_multi_line_comment,
            self._parse_variable_definition,
            self._parse_command,
            self._parse_title,
            self._parse_attributes,
            self._parse_header,
            self._parse_block,
            self._parse_content,
            self._parse_list,
            self._parse_paragraph,
        ]

    def _create_toc(self):
        # Create the TOC from the list of headers.

        nodes = []
        latest_by_level = {}

        for header_node in self.headers:
            # This is the current node
            node = TocEntryNode(header_node)
            level = header_node.level

            # This collects the latest node added with a given level
            latest_by_level[level] = node

            try:
                # Simplest case, add it to the latest one
                # with a level just 1 step lower
                latest_by_level[level - 1].children.append(node)
            except KeyError:
                # Find all the latest ones added with a level lower than this
                latest = [latest_by_level.get(i, None) for i in range(1, level)]

                # Get the children list of each one, plus nodes for the root
                children = [nodes] + [i.children for i in latest if i is not None]

                # Get the nearest one and append to that
                children[-1].append(node)

        return TocNode(entries=nodes)

    def parse(self):
        super().parse()

        self.toc = self._create_toc()
        self.footnotes = FootnotesNode(entries=self.footnote_defs)



from mau import text_buffer
from mau.lexers.base_lexer import BaseLexer, Token, TokenTypes


class PreprocessVariablesLexer(BaseLexer):
    def _process_eol(self):
        try:
            self._current_char
        except text_buffer.EOLError:
            self._nextline()
            return Token(TokenTypes.TEXT, "\n")

    def _process_literal(self):
        if self._current_char not in r"\`{}":
            return None

        return self._create_token_and_skip(TokenTypes.LITERAL, self._current_char)

    def _process_text(self):
        return self._create_token_and_skip(TokenTypes.TEXT, self._current_char)

    def _process_functions(self):
        return [
            self._process_eof,
            self._process_eol,
            self._process_literal,
            self._process_text,
        ]

    def process(self, text):
        super().process(text)

        self.tokens = self.tokens[:-1]



class EOLError(ValueError):

    """ Signals that the buffer is reading after the end of a line."""


class EOFError(ValueError):

    """ Signals that the buffer is reading after the end of the text."""


class TextBuffer:
    def __init__(self, text=None):
        self.lines = []
        self.reset()

        if text is not None:
            self.load(text)

    def reset(self):
        self.line = 0
        self.column = 0

    def load(self, text):
        self.lines = text.split("\n") if text is not None else []
        self.reset()

    @property
    def current_line(self):
        """
        Returns the current line.

        This property returns the current line without advancing the index.
        If the buffer is reading after the last line it raises EOFError.
        """
        try:
            return self.lines[self.line]
        except IndexError:
            raise EOFError("EOF reading line {}".format(self.line))

    @property
    def current_char(self):
        """
        Returns the current character.

        This property returns the current character without advancing the index.
        If the buffer is reading after the last character of the line it raises EOLError.
        """
        try:
            return self.current_line[self.column]
        except IndexError:
            raise EOLError(
                "EOL reading column {} at line {}".format(self.column, self.line)
            )

    @property
    def peek_char(self):
        """
        Returns the next character.

        This property returns the next character without advancing the index.
        If the buffer is reading after the last character of the line it raises EOLError.
        """

        try:
            return self.current_line[self.column + 1]
        except IndexError:
            raise EOLError(
                "EOL reading column {} at line {}".format(self.column, self.line)
            )

    @property
    def tail(self):
        """
        Returns the remaining part of the line.

        This property returns a string with the last part of the current
        line from the current character to the end.
        """
        return self.current_line[self.column :]

    @property
    def position(self):
        """
        Returns a tuple with the current position.
        """
        return (self.line, self.column)

    @position.setter
    def position(self, position):
        """
        Sets the current position.
        """
        self.line, self.column = position

    def nextline(self):
        """
        Moves the index to the beginning of the next line
        """
        self.line += 1
        self.column = 0

    def skip(self, chars=1):
        """
        Skips the given number of characters (default 1). Can silently
        go over the end of the line.
        """
        self.column += chars

    def goto(self, line, column=0):
        self.line, self.column = line, column

    def context(self, line_number, column_number):
        text_line = self.lines[line_number]

        prefix = f"{line_number}: "
        suffix = ""

        HALF_SIZE = 32
        min_column = max(0, column_number - HALF_SIZE)
        max_column = min(len(text_line), column_number + HALF_SIZE + 1)

        if min_column > 0:
            prefix = f"{prefix}[...]"

        if max_column < len(text_line):
            suffix = "[...]"

        text_line = text_line[min_column:max_column]

        # Create the caret line
        caret_line = " " * (column_number - min_column + len(prefix)) + "^"

        return {
            "text": [f"{prefix}{text_line}{suffix}", caret_line],
            "line": line_number,
            "column": column_number,
        }

    def insert(self, text):
        """
        Inserts the given text in the buffer.
        """

        # Split the new text removing the trailing newline
        splittext = text.strip().split("\n") if text is not None else []

        # Insert the new text lines into the current buffer
        newlines = self.lines[: self.line] + splittext + self.lines[self.line :]

        self.lines = newlines



import re
import string
from functools import partial
from collections.abc import Sequence

from mau import text_buffer


class TokenTypes:
    EOL = "EOL"
    EOF = "EOF"
    LITERAL = "LITERAL"
    TEXT = "TEXT"
    WHITESPACE = "WHITESPACE"


class LexerError(ValueError):
    pass


class Token:
    """
    This represents a token.
    Tokens have a type, a value (the actual characters), and a position
    in the global text, expressed as a tuple of line and column
    """

    def __init__(self, _type, value=None, position=None):
        self.type = _type
        self.position = position

        # This ensures things like numbers or other
        # strange beasts are just treated as text.
        self.value = str(value) if value is not None else None

    def __str__(self):
        position_string = ""
        if self.position:
            position_string = f", line={self.position[0]}, col={self.position[1]}"

        value_string = ""
        if self.value is not None:
            value_string = f", '{self.value}'"

        return f"Token({self.type}{value_string}{position_string})"

    __repr__ = __str__

    def __eq__(self, other):
        if other.value is None:
            return self.type == other.type

        return (self.type, self.value) == (
            other.type,
            other.value,
        )

    def __hash__(self):
        return hash((self.type, self.value))

    def __len__(self):
        if self.value:
            return len(self.value)

        return 0

    def __bool__(self):
        return True


# These are convenient shortcuts
EOL = Token(TokenTypes.EOL)
EOF = Token(TokenTypes.EOF)
WS = partial(Token, TokenTypes.WHITESPACE)
Text = partial(Token, TokenTypes.TEXT)
Literal = partial(Token, TokenTypes.LITERAL)


class BaseLexer:
    """
    The base class for lexers.
    The lexer decomposes the input text into a list of tokens
    and provides basic navigation functions in the
    output results.
    """

    def __init__(self, initial_position=None):
        # Use a TextBuffer internally to manage the text
        self._text_buffer = text_buffer.TextBuffer()

        # A buffer of tokens, useful when you need to
        # collect them but later post-process them
        # before you actually store them as result.
        self._buffer = []

        self._initial_position = initial_position or (0, 0)

        # These are the tokens identified so far
        self.tokens = []

    @property
    def _token_position(self):
        # This returns the token position taking into
        # account that the initial position might
        # not be (0,0)
        return tuple(map(sum, zip(self._text_buffer.position, self._initial_position)))

    def context(self, token):
        """Returns the context of a token for error reporting purposes"""
        return self._text_buffer.context(*token.position)

    def process(self, text):
        # Reset the lexer
        self.tokens = []

        # Load the text into the TextBuffer
        self._text_buffer.load(text)

        # Process tokens until we reach the end of file
        self._process()
        while True:
            # Preprocess functions can return no tokens
            if len(self.tokens) > 0 and self.tokens[-1].type is TokenTypes.EOF:
                break

            self._process()

    def _process(self):
        # This should not be touched by child classes
        # as it is the core of the lexer. It tries
        # each function in the list returned by
        # _process_functions and stores all the resulting
        # tokens. A parsing function must return None
        # when characters do not match the rules.

        process_functions = self._process_functions()
        process_functions.append(self._process_error)

        for process_func in process_functions:
            # This ensures result is always either None or a list
            result = self._wrap(process_func())

            if result is None:
                continue

            self.tokens.extend(result)
            return

    def _wrap(self, result):
        # Makes sure the result is either None or a list of tokens
        # which makes  processing function that return a single token
        # more readable.

        if result is None:
            return

        if not isinstance(result, Sequence):
            return [result]

        return result

    def _nextline(self):
        # Carriage return =) go to column 0
        self._initial_position = (self._initial_position[0], 0)

        # Skip the whole line including the EOL
        self._text_buffer.nextline()

    def _skip(self, steps=1):
        # Skip only the given amount of characters
        self._text_buffer.skip(steps)

    @property
    def _current_char(self):
        # Return the current character
        return self._text_buffer.current_char

    @property
    def _current_line(self):
        # Return the current line
        return self._text_buffer.current_line

    @property
    def _tail(self):
        # A wrapper to return the rest of the line
        return self._text_buffer.tail

    def _create_token(self, token_type, token_value=None):
        # A wrapper to create a token with the current position.
        # This doesn't affect the position in the text being lexed
        return Token(token_type, token_value, position=self._token_position)

    def _create_token_and_skip(self, token_type, token_value=None, skip_value=None):
        # Create the token and skip the characters in the text

        # This skips the first non-None value between skip_value, token_type and the empty string
        skip = next(x for x in [skip_value, token_value, ""] if x is not None)

        # This creates the token
        token = self._create_token(token_type, token_value)

        # Perform the right skip
        if token_type == TokenTypes.EOL:
            self._nextline()
        else:
            self._skip(len(skip))

        return token

    def _store(self, token_type, token_value=None, skip_value=None):
        # Create and skip a token, then store it in the buffer
        self._buffer.append(
            self._create_token_and_skip(token_type, token_value, skip_value)
        )

    def _pop(self):
        # Get the content of the buffer
        tokens = list(self._buffer)
        self._buffer = []
        return tokens

    def _process_eof(self):
        try:
            self._current_line
        except text_buffer.EOFError:
            return self._create_token_and_skip(TokenTypes.EOF)

    def _process_eol(self):
        try:
            self._current_char
        except text_buffer.EOLError:
            return self._create_token_and_skip(TokenTypes.EOL)

    def _process_character(self):
        if self._current_char not in string.ascii_letters:
            return None

        self._store(TokenTypes.TEXT, self._current_char)
        return self._pop()

    def _process_whitespace(self):
        regexp = re.compile(r"\ +")

        match = regexp.match(self._tail)

        if not match:
            return None

        self._store(TokenTypes.WHITESPACE, match.group())
        return self._pop()

    def _process_functions(self):
        return [
            self._process_eof,
            self._process_eol,
            self._process_whitespace,
            self._process_character,
        ]

    def _process_error(self):
        raise LexerError(f'Can\'t process "{self._tail}"')

    def _insert(self, text):
        self._text_buffer.insert(text)

    def _rematch(self, regexp):
        # Compile the regexp and get a match on the current line
        regexp = re.compile(regexp)
        return regexp.match(self._current_line)



# -*- coding: utf-8 -*-
from pkg_resources import DistributionNotFound, get_distribution

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound

# This is not used here but exported to __init__.py
from mau.parsers.base_parser import ParserError
from mau.parsers.main_parser import MainParser
from mau.parsers.nodes import DocumentNode, ContainerNode
from mau.visitors.html_visitor import HTMLVisitor
from mau.visitors.asciidoctor_visitor import AsciidoctorVisitor
from mau.visitors.markua_visitor import MarkuaVisitor


class ConfigurationError(ValueError):
    """ Used to signal an error in the configuration """


class Mau:
    def __init__(
        self,
        config,
        target_format,
        default_templates=None,
        custom_templates=None,
        templates_directory=None,
        full_document=False,
    ):
        # This is an external configuration file that
        # will be used here and also injected into the variables
        self.config = config

        # A dictionary with the default templates.
        self.default_templates = default_templates

        # A dictionary with the custom templates.
        self.custom_templates = custom_templates

        # A directory that contains the custom templates as files
        self.templates_directory = templates_directory

        # If this flag is True the output is wrapped in a DocumentNode,
        # otherwise Mau will use a ContainerNode
        self.full_document = full_document

        # This will contain all the variables declared
        # in the text and in the configuration
        self.variables = {}

    def process(self, text):
        # Store the config under the Mau namespace
        self.variables["mau"] = self.config

        # Parse the source text using the given configuration
        parser = MainParser(variables=self.variables)
        parser.load(text)
        parser.parse()

        wrapper_node_class = ContainerNode
        if self.full_document:
            wrapper_node_class = DocumentNode

        # Wrap the whole output
        output = wrapper_node_class(parser.nodes)

        # Select the visitor class
        target_format = self.config.get("target_format", "html")

        if target_format == "html":
            visitor_class = HTMLVisitor
        elif target_format == "asciidoctor":
            visitor_class = AsciidoctorVisitor
        elif target_format == "markua":
            visitor_class = MarkuaVisitor
        else:
            raise ConfigurationError(f"Target format {target_format} is not available")

        # Initialise the visitor
        # Use the parser variables so that the visitor
        # has both the configuration values and the
        # variables defined inside the text
        visitor = visitor_class(
            default_templates=self.default_templates,
            custom_templates=self.custom_templates,
            templates_directory=self.templates_directory,
            config=parser.variables,
            toc=parser.toc,
            footnotes=parser.footnotes,
        )

        # Replace variables with the ones processed by the parser
        self.variables = parser.variables

        # Visit the document AST
        return visitor.visit(output.asdict())



from mau.lexers.base_lexer import Text, Literal, EOL, EOF
from mau.lexers.text_lexer import TextLexer


def test_normal_text():
    lex = TextLexer()

    lex.process("Normal text")

    assert lex.tokens == [
        Text("Normal"),
        Text(" "),
        Text("text"),
        EOL,
        EOF,
    ]


def test_underscore():
    lex = TextLexer()

    lex.process("_underscore_")

    assert lex.tokens == [
        Literal("_"),
        Text("underscore"),
        Literal("_"),
        EOL,
        EOF,
    ]


def test_star():
    lex = TextLexer()

    lex.process("*star*")

    assert lex.tokens == [
        Literal("*"),
        Text("star"),
        Literal("*"),
        EOL,
        EOF,
    ]


def test_backtick():
    lex = TextLexer()

    lex.process("`backtick`")

    assert lex.tokens == [
        Literal("`"),
        Text("backtick"),
        Literal("`"),
        EOL,
        EOF,
    ]


def test_curly_braces():
    lex = TextLexer()

    lex.process("{curly}")

    assert lex.tokens == [
        Literal("{"),
        Text("curly"),
        Literal("}"),
        EOL,
        EOF,
    ]


def test_round_brackets():
    lex = TextLexer()

    lex.process("(round)")

    assert lex.tokens == [
        Literal("("),
        Text("round"),
        Literal(")"),
        EOL,
        EOF,
    ]


def test_square_brackets():
    lex = TextLexer()

    lex.process("[square]")

    assert lex.tokens == [
        Literal("["),
        Text("square"),
        Literal("]"),
        EOL,
        EOF,
    ]


def test_single_class():
    lex = TextLexer()

    lex.process("[name]#content#")

    assert lex.tokens == [
        Literal("["),
        Text("name"),
        Literal("]"),
        Literal("#"),
        Text("content"),
        Literal("#"),
        EOL,
        EOF,
    ]


def test_multiple_classes():
    lex = TextLexer()

    lex.process("[name1,name2]#content#")

    assert lex.tokens == [
        Literal("["),
        Text("name1,name2"),
        Literal("]"),
        Literal("#"),
        Text("content"),
        Literal("#"),
        EOL,
        EOF,
    ]


def test_class_content_with_styles():
    lex = TextLexer()

    lex.process("[name]#_content_#")

    assert lex.tokens == [
        Literal("["),
        Text("name"),
        Literal("]"),
        Literal("#"),
        Literal("_"),
        Text("content"),
        Literal("_"),
        Literal("#"),
        EOL,
        EOF,
    ]


def test_macro():
    lex = TextLexer()

    lex.process("[macro](value1,value2)")

    assert lex.tokens == [
        Literal("["),
        Text("macro"),
        Literal("]"),
        Literal("("),
        Text("value1,value2"),
        Literal(")"),
        EOL,
        EOF,
    ]


def test_macro_named_attributes():
    lex = TextLexer()

    lex.process("[macro](attr1=value1,attr2=value2)")

    assert lex.tokens == [
        Literal("["),
        Text("macro"),
        Literal("]"),
        Literal("("),
        Text("attr1=value1,attr2=value2"),
        Literal(")"),
        EOL,
        EOF,
    ]


def test_escaped_underscore():
    lex = TextLexer()

    lex.process(r"\_underscore\_")

    # Escaped characters are TEXT tokens, not LITERAL
    assert lex.tokens == [
        Literal("\\"),
        Literal("_"),
        Text("underscore"),
        Literal("\\"),
        Literal("_"),
        EOL,
        EOF,
    ]


def test_escaped_square_brackets():
    lex = TextLexer()

    lex.process(r"\[square\]")

    # Escaped characters are TEXT tokens, not LITERAL
    assert lex.tokens == [
        Literal("\\"),
        Literal("["),
        Text("square"),
        Literal("\\"),
        Literal("]"),
        EOL,
        EOF,
    ]


def test_escaped_round_brackets():
    lex = TextLexer()

    lex.process(r"\(round\)")

    # Escaped characters are TEXT tokens, not LITERAL
    assert lex.tokens == [
        Literal("\\"),
        Literal("("),
        Text("round"),
        Literal("\\"),
        Literal(")"),
        EOL,
        EOF,
    ]


def test_escaped_curly_braces():
    lex = TextLexer()

    lex.process(r"\{curly\}")

    # Escaped characters are TEXT tokens, not LITERAL
    assert lex.tokens == [
        Literal("\\"),
        Literal("{"),
        Text("curly"),
        Literal("\\"),
        Literal("}"),
        EOL,
        EOF,
    ]


def test_escaped_quotes():
    lex = TextLexer()

    lex.process(r"\"quotes\"")

    assert lex.tokens == [
        Literal("\\"),
        Literal('"'),
        Text("quotes"),
        Literal("\\"),
        Literal('"'),
        EOL,
        EOF,
    ]



import copy

from mau.visitors.visitor import Visitor

DEFAULT_TEMPLATES = {
    "block-admonition.adoc": "[{{ kwargs.class }}{% if kwargs.icon %}.{{ kwargs.icon }}{% endif %}]\n====\n{{ content }}====\n",
    "block.adoc": "--\n{{ content }}\n--",
    "block-source.adoc": "{% if title %}.{{ title }}\n{% endif %}[source{% if kwargs.language %},{{ kwargs.language }}{% endif %}]\n----\n{{ content }}\n----\n{% if kwargs.callouts %}{% for callout in kwargs.callouts %}{{ callout[0] }} {{ callout[1] }}{% endfor %}\n{% endif %}",
    "callout.adoc": "<{{ name }}>",
    "class.adoc": """[{{ classes }}]#{{ content }}#""",
    "command.adoc": "{{ content }}",
    "document.adoc": "{{ content }}",
    "footnote_ref.adoc": "footnote:[{{ content }}]",
    "header.adoc": "{{ header }} {{ value }}\n",
    "horizontal_rule.adoc": "---\n",
    "image.adoc": "{% if asciidoctor_classes %}[{{ asciidoctor_classes }}]\n{% endif %}{% if title %}.{{ title }}\n{% endif %}image::{{ uri }}[{% if alt_text %}{{ alt_text }}{% endif %}]\n",
    "inline_image.adoc": "image::{{ uri }}{%if alt_text %}[{{ alt_text }}]{% endif %}",
    "link.adoc": "{{ target }}{% if text %}[{{ text }}]{% endif %}",
    "list.adoc": "{{ items }}{% if main_node %}\n{% endif %}",
    "list_item.adoc": "{% if not main_node %}\n{% endif %}{% if prefix %}{{ prefix }} {% endif %}{{ content }}",
    "paragraph.adoc": "{{ content }}\n",
    "block-quote.adoc": '[quote, "{% if kwargs.attribution %}{{ kwargs.attribution }}{% else %}{{ secondary_content }}{% endif %}"]\n____\n{{ content }}____\n',
    "sentence.adoc": "{{ content }}",
    "star.adoc": "*{{ content }}*",
    "text.adoc": "{{ value }}",
    "underscore.adoc": "_{{ content }}_",
    "verbatim.adoc": "`{{ content }}`",
}


class AsciidoctorVisitor(Visitor):
    _template_extension = ".adoc"
    _environment_options = {"keep_trailing_newline": True}

    def __init__(
        self,
        default_templates=DEFAULT_TEMPLATES,
        custom_templates=None,
        templates_directory=None,
        config=None,
        toc=None,
        footnotes=None,
    ):
        default_templates = (
            copy.deepcopy(default_templates)
            if default_templates is not None
            else copy.deepcopy(DEFAULT_TEMPLATES)
        )

        super().__init__(
            default_templates=default_templates,
            custom_templates=custom_templates,
            templates_directory=templates_directory,
            config=config,
            toc=toc,
            footnotes=footnotes,
        )

    def _visit_class(self, node):
        classes = " ".join([f".{cls}" for cls in node["classes"]])
        node["classes"] = classes
        return super()._visit_class(node)

    def _visit_link(self, node):
        if node["text"] == node["target"]:
            node["text"] = None
        return node

    def _visit_header(self, node):
        node["header"] = "=" * int(node["level"])
        return node

    def _visit_block_admonition(self, node):
        if node["kwargs"]["class"] in [
            "note",
            "tip",
            "important",
            "caution",
            "warning",
        ]:
            node["kwargs"]["class"] = node["kwargs"]["class"].upper()
        else:
            raise ValueError(
                f"""Admonition {node["kwargs"]["class"]} cannot be converted"""
            )

        return node

    def _visit_block_engine_source(self, node):
        src = node["content"].split("\n")

        # Inject callout markers in the highlighted code
        callout_markers = node["kwargs"]["callouts"]["markers"]
        callout_contents = node["kwargs"]["callouts"]["contents"]

        for linenum, callout_name in callout_markers.items():
            src[linenum] = "{line} {callout}".format(
                line=src[linenum],
                callout=self._render("callout", name=callout_name),
            )

        callouts_list = [
            (self._render("callout", name=name), text)
            for name, text in callout_contents.items()
        ]

        node["content"] = "\n".join(src)
        node["kwargs"]["callouts"] = callouts_list

        return node

    def _visit_document(self, node):
        self._reducelist(node, ["content"], join_with="\n")
        return node

    def _visit_list_item(self, node, ordered=False):
        mark = "*"
        if ordered:
            mark = "."

        prefix = None
        if node["content"]["type"] != "list":
            prefix = mark * int(node["level"])

        self._reduce(node, ["content"])
        node["prefix"] = prefix
        return node

    def _visit_list(self, node, ordered=False):
        node["items"] = "".join(
            [self.visit(i, ordered=node["ordered"]) for i in node["items"]]
        )
        return node

    def _visit_footnote_ref(self, node):
        node["node_types"] = (["footnote_ref"],)
        self._reducelist(node, ["content"], join_with="")
        return node

    def _visit_content_image(self, node):
        node["node_types"] = ["image"]
        node["asciidoctor_classes"] = node["kwargs"].get("asciidoctor_classes", None)
        self._reduce(node, ["title"])
        return node

    def _visit_image(self, node):
        node["node_types"] = ["inline_image"]
        return node



from mau.lexers.base_lexer import Token, TokenTypes, Text, Literal, EOL, EOF
from mau.lexers.preprocess_variables_lexer import PreprocessVariablesLexer


def test_normal_text():
    lex = PreprocessVariablesLexer()

    lex.process("Some text")

    assert lex.tokens == [
        Text("S"),
        Text("o"),
        Text("m"),
        Text("e"),
        Text(" "),
        Text("t"),
        Text("e"),
        Text("x"),
        Text("t"),
        Text("\n"),
    ]


def test_match_only_backticks_and_curly_braces():
    lex = PreprocessVariablesLexer()

    lex.process("Normal text `{curly}` _other_ *text*")

    assert lex.tokens == [
        Text("N"),
        Text("o"),
        Text("r"),
        Text("m"),
        Text("a"),
        Text("l"),
        Text(" "),
        Text("t"),
        Text("e"),
        Text("x"),
        Text("t"),
        Text(" "),
        Literal("`"),
        Literal("{"),
        Text("c"),
        Text("u"),
        Text("r"),
        Text("l"),
        Text("y"),
        Literal("}"),
        Literal("`"),
        Text(" "),
        Text("_"),
        Text("o"),
        Text("t"),
        Text("h"),
        Text("e"),
        Text("r"),
        Text("_"),
        Text(" "),
        Text("*"),
        Text("t"),
        Text("e"),
        Text("x"),
        Text("t"),
        Text("*"),
        Text("\n"),
    ]


def test_escape_curly_braces():
    lex = PreprocessVariablesLexer()

    lex.process(r"Normal text \{curly\} _other_ *text*")

    assert lex.tokens == [
        Text("N"),
        Text("o"),
        Text("r"),
        Text("m"),
        Text("a"),
        Text("l"),
        Text(" "),
        Text("t"),
        Text("e"),
        Text("x"),
        Text("t"),
        Text(" "),
        Literal("\\"),
        Literal("{"),
        Text("c"),
        Text("u"),
        Text("r"),
        Text("l"),
        Text("y"),
        Literal("\\"),
        Literal("}"),
        Text(" "),
        Text("_"),
        Text("o"),
        Text("t"),
        Text("h"),
        Text("e"),
        Text("r"),
        Text("_"),
        Text(" "),
        Text("*"),
        Text("t"),
        Text("e"),
        Text("x"),
        Text("t"),
        Text("*"),
        Text("\n"),
    ]


def test_preserve_escapes():
    lex = PreprocessVariablesLexer()

    lex.process(r"Normal \text \_other\_")

    assert lex.tokens == [
        Text("N"),
        Text("o"),
        Text("r"),
        Text("m"),
        Text("a"),
        Text("l"),
        Text(" "),
        Literal("\\"),
        Text("t"),
        Text("e"),
        Text("x"),
        Text("t"),
        Text(" "),
        Literal("\\"),
        Text("_"),
        Text("o"),
        Text("t"),
        Text("h"),
        Text("e"),
        Text("r"),
        Literal("\\"),
        Text("_"),
        Text("\n"),
    ]



from unittest.mock import patch

from mau.parsers import nodes
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_default_header_anchor_function():
    source = """
    = Some Words 1234 56
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Some Words 1234 56",
            "level": 1,
            "anchor": "some-words-1234-56",
        }
    ]

    _test(source, expected)


def test_default_header_anchor_function_multiple_spaces():
    source = """
    = Some    Words     1234    56
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Some    Words     1234    56",
            "level": 1,
            "anchor": "some-words-1234-56",
        }
    ]

    _test(source, expected)


def test_default_header_anchor_function_filter_characters():
    source = """
    = Some #Words @ 12!34 56
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Some #Words @ 12!34 56",
            "level": 1,
            "anchor": "some-words-1234-56",
        }
    ]

    _test(source, expected)


def test_custom_header_anchor_function():
    source = """
    = Title of the section
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Title of the section",
            "level": 1,
            "anchor": "XXXXXY",
        }
    ]

    config = {"mau": {"header_anchor_function": lambda text, level: "XXXXXY"}}

    _test = parser_test_factory(MainParser, variables=config)

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_header_level_1(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = """
    = Title of the section
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Title of the section",
            "level": 1,
            "anchor": "XXXXXX",
        }
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_header_level_3(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = """
    === Title of a subsection
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Title of a subsection",
            "level": 3,
            "anchor": "XXXXXX",
        }
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_collect_headers(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    = Header 1
    == Header 1.1
    == Header 1.2
    = Header 2
    == Header 2.1
    === Header 2.1.1
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1",
            "level": 1,
            "anchor": "Header 1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1.1",
            "level": 2,
            "anchor": "Header 1.1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1.2",
            "level": 2,
            "anchor": "Header 1.2-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 2",
            "level": 1,
            "anchor": "Header 2-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 2.1",
            "level": 2,
            "anchor": "Header 2.1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 2.1.1",
            "level": 3,
            "anchor": "Header 2.1.1-XXXXXX",
        },
    ]

    p = _test(source, expected)

    assert p.headers == [
        nodes.HeaderNode("Header 1", 1, "Header 1-XXXXXX"),
        nodes.HeaderNode("Header 1.1", 2, "Header 1.1-XXXXXX"),
        nodes.HeaderNode("Header 1.2", 2, "Header 1.2-XXXXXX"),
        nodes.HeaderNode("Header 2", 1, "Header 2-XXXXXX"),
        nodes.HeaderNode("Header 2.1", 2, "Header 2.1-XXXXXX"),
        nodes.HeaderNode("Header 2.1.1", 3, "Header 2.1.1-XXXXXX"),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_attributes_header(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    [value1,someattr1=somevalue1,someattr2=somevalue2]
    = Header
    """

    expected = [
        {
            "type": "header",
            "value": "Header",
            "level": 1,
            "kwargs": {"someattr1": "somevalue1", "someattr2": "somevalue2"},
            "tags": [],
            "anchor": "Header-XXXXXX",
        },
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_single_tag_header(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    [tags=section]
    = Header
    """

    expected = [
        {
            "type": "header",
            "value": "Header",
            "level": 1,
            "kwargs": {},
            "tags": ["section"],
            "anchor": "Header-XXXXXX",
        },
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_multiple_tags_header(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    [tags="section,important"]
    = Header
    """

    expected = [
        {
            "type": "header",
            "value": "Header",
            "level": 1,
            "kwargs": {},
            "tags": ["section", "important"],
            "anchor": "Header-XXXXXX",
        },
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_attributes_and_tags_header(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    [value1,someattr1=somevalue1,someattr2=somevalue2,tags="section,important"]
    = Header
    """

    expected = [
        {
            "type": "header",
            "value": "Header",
            "level": 1,
            "kwargs": {"someattr1": "somevalue1", "someattr2": "somevalue2"},
            "tags": ["section", "important"],
            "anchor": "Header-XXXXXX",
        },
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_headers_not_in_toc(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    = Header 1
    == Header 1.1
    ==! Header 1.2
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1",
            "level": 1,
            "anchor": "Header 1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1.1",
            "level": 2,
            "anchor": "Header 1.1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1.2",
            "level": 2,
            "anchor": "Header 1.2-XXXXXX",
        },
    ]

    p = _test(source, expected)

    assert p.headers == [
        nodes.HeaderNode("Header 1", 1, "Header 1-XXXXXX"),
        nodes.HeaderNode("Header 1.1", 2, "Header 1.1-XXXXXX"),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_headers_not_in_toc_with_children(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    = Header 1
    == Header 1.1
    ==! Header 1.2
    === Header 1.2.1
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1",
            "level": 1,
            "anchor": "Header 1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1.1",
            "level": 2,
            "anchor": "Header 1.1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1.2",
            "level": 2,
            "anchor": "Header 1.2-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "tags": [],
            "value": "Header 1.2.1",
            "level": 3,
            "anchor": "Header 1.2.1-XXXXXX",
        },
    ]

    p = _test(source, expected)

    assert p.headers == [
        nodes.HeaderNode("Header 1", 1, "Header 1-XXXXXX"),
        nodes.HeaderNode("Header 1.1", 2, "Header 1.1-XXXXXX"),
        nodes.HeaderNode("Header 1.2.1", 3, "Header 1.2.1-XXXXXX"),
    ]



import textwrap
from unittest.mock import patch

import pytest

from mau.parsers.main_parser import MainParser
from mau.visitors.visitor import CommandError
from mau.visitors.html_visitor import HTMLVisitor

from tests.helpers import (
    init_parser_factory,
    init_ast_factory,
    ast_test_factory,
)

init_parser = init_parser_factory(MainParser)

init_ast = init_ast_factory(MainParser)

_test = ast_test_factory(MainParser, HTMLVisitor)


@patch("mau.visitors.html_visitor.highlight")
def test_source(mock_highlight):
    mock_highlight.return_value = "XYXYXY"

    source = textwrap.dedent(
        """
        [myblock, engine=source]
        ----
        import os

        print(os.environ["HOME"])
        ----
        """
    )

    expected = ['<div class="myblock"><div class="content">XYXYXY</div></div>']

    _test(source, expected)


@patch("mau.visitors.html_visitor.highlight")
def test_source_indentation(mock_highlight):
    mock_highlight.side_effect = lambda src, lexer, formatter: src

    source = textwrap.dedent(
        """
        [myblock, engine=source]
        ----
        Not indented

            Indented
        ----
        """
    )

    expected = [
        '<div class="myblock"><div class="content">Not indented\n\n    Indented</div></div>'
    ]

    _test(source, expected)


def test_source_callouts():
    source = textwrap.dedent(
        """
        [myblock, engine=source]
        ----
        Some code:1:
        
        Some other code:3:
        ----
        1: This is the first line
        3: This is the second line
        """
    )

    expected = [
        (
            '<div class="myblock">'
            '<div class="content">'
            '<div class="highlight">'
            '<pre><span></span>Some code <span class="callout">1</span>\n\nSome other code <span class="callout">3</span>\n</pre>'
            "</div>\n</div>"
            '<div class="callouts">'
            "<table><tbody>"
            '<tr><td><span class="callout">1</span></td><td>This is the first line</td></tr>'
            '<tr><td><span class="callout">3</span></td><td>This is the second line</td></tr>'
            "</tbody></table>"
            "</div></div>"
        )
    ]

    _test(source, expected)


def test_source_empty_callouts():
    source = textwrap.dedent(
        """
        [myblock, engine=source]
        ----
        Some code:1:
        
        Some other code:3:
        ----
        """
    )

    expected = [
        (
            '<div class="myblock">'
            '<div class="content">'
            '<div class="highlight">'
            '<pre><span></span>Some code <span class="callout">1</span>\n\nSome other code <span class="callout">3</span>\n</pre>'
            "</div>\n</div>"
            "</div>"
        )
    ]

    _test(source, expected)


def test_source_pygments_config():
    source = textwrap.dedent(
        """
        [myblock, engine=source, pygments.hl_lines="1,2,3"]
        ----
        import os

        print(os.environ["HOME"])
        ----
        """
    )

    expected = [
        '<div class="myblock"><div class="content"><div class="highlight"><pre><span></span><span class="hll">import os\n</span><span class="hll">\n</span><span class="hll">print(os.environ[&quot;HOME&quot;])\n</span></pre></div>\n</div></div>'
    ]

    _test(source, expected)


def test_source_highlights():
    source = textwrap.dedent(
        """
        [myblock, engine=source]
        ----
        import os:@:

        print(os.environ["HOME"]):@:
        ----
        """
    )

    expected = [
        '<div class="myblock"><div class="content"><div class="highlight"><pre><span></span><span class="hll">import os\n</span>\n<span class="hll">print(os.environ[&quot;HOME&quot;])\n</span></pre></div>\n</div></div>'
    ]

    _test(source, expected)


@patch("mau.visitors.html_visitor.highlight")
def test_source_block_definition(mock_highlight):
    mock_highlight.return_value = "XYXYXY"

    source = textwrap.dedent(
        """
        ::defblock:source, myblock, engine=source

        [source]
        ----
        import os

        print(os.environ["HOME"])
        ----
        """
    )

    expected = ['<div class="myblock"><div class="content">XYXYXY</div></div>']

    _test(source, expected)


@patch("mau.visitors.html_visitor.highlight")
def test_source_block_definition_variables_are_untouched(mock_highlight):
    mock_highlight.return_value = "XYXYXY"

    source = textwrap.dedent(
        """
        [source]
        ----
        a = 5

        print(f"{a}")
        ----
        """
    )

    expected = ['<div class="source"><div class="content">XYXYXY</div></div>']

    _test(source, expected)



import textwrap

import pytest

from tests.helpers import dedent

from mau import text_buffer


def test_text_buffer_init_empty():
    tb = text_buffer.TextBuffer()

    with pytest.raises(text_buffer.EOFError):
        tb.current_char


def test_text_buffer_init_one_line():
    tb = text_buffer.TextBuffer("abcdef")

    assert tb.current_char == "a"
    assert tb.peek_char == "b"
    assert tb.current_line == "abcdef"
    assert tb.line == 0
    assert tb.column == 0


def test_text_buffer_end_of_line_peek_char():
    tb = text_buffer.TextBuffer("abcdef")
    tb.column = 5

    assert tb.current_char == "f"
    with pytest.raises(text_buffer.EOLError):
        tb.peek_char


def test_text_buffer_end_of_line_current_char():
    tb = text_buffer.TextBuffer("abcdef")
    tb.column = 200

    with pytest.raises(text_buffer.EOLError):
        tb.current_char


def test_text_buffer_error_at_end_of_file():
    tb = text_buffer.TextBuffer("abcdef")
    tb.line = 1

    with pytest.raises(text_buffer.EOFError):
        tb.current_line


def test_text_buffer_error_after_end_of_file():
    tb = text_buffer.TextBuffer("abcdef")
    tb.line = 100

    with pytest.raises(text_buffer.EOFError):
        tb.current_line


def test_text_buffer_tail():
    ts = text_buffer.TextBuffer("abcdefgh")
    ts.column = 4

    assert ts.tail == "efgh"


def test_text_buffer_multiple_lines():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.line = 1
    tb.column = 1

    assert tb.current_line == "def"
    assert tb.current_char == "e"
    assert tb.peek_char == "f"


def test_text_buffer_nextline():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.line = 1
    tb.column = 2
    tb.nextline()

    assert tb.position == (2, 0)


def test_text_buffer_position():
    tb = text_buffer.TextBuffer()
    tb.line = 123
    tb.column = 456

    assert tb.position == (123, 456)


def test_text_buffer_set_position():
    tb = text_buffer.TextBuffer()
    tb.position = (222, 333)

    assert tb.position == (222, 333)


def test_text_buffer_skip_defaults_to_one():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.skip()

    assert tb.column == 1


def test_text_buffer_skip_accepts_value():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.skip(3)

    assert tb.column == 3


def test_text_buffer_goto():
    tb = text_buffer.TextBuffer()
    tb.goto(12, 45)

    assert tb.position == (12, 45)


def test_text_buffer_goto_default_column():
    tb = text_buffer.TextBuffer()
    tb.goto(12)

    assert tb.position == (12, 0)


def test_context():
    tb = text_buffer.TextBuffer(
        dedent(
            """
            Some text
            on multiple lines
            that I will use
            to check the context()
            """
        )
    )

    assert tb.context(2, 7) == {
        "text": [
            "2: that I will use",
            "          ^",
        ],
        "line": 2,
        "column": 7,
    }


def test_context_long_line_center():
    tb = text_buffer.TextBuffer(
        dedent(
            """
            Some text on a very long line that I will use to check the context() and to test that the whole thing works
            """
        )
    )

    assert tb.context(0, 50) == {
        "text": [
            "0: [...]y long line that I will use to check the context() and to test th[...]",
            "                                        ^",
        ],
        "line": 0,
        "column": 50,
    }


def test_context_long_line_beginning():
    tb = text_buffer.TextBuffer(
        dedent(
            """
            Some text on a very long line that I will use to check the context() and to test that the whole thing works
            """
        )
    )

    assert tb.context(0, 30) == {
        "text": [
            "0: Some text on a very long line that I will use to check the cont[...]",
            "                                 ^",
        ],
        "line": 0,
        "column": 30,
    }


def test_context_long_line_end():
    tb = text_buffer.TextBuffer(
        dedent(
            """
            Some text on a very long line that I will use to check the context() and to test that the whole thing works
            """
        )
    )

    assert tb.context(0, 100) == {
        "text": [
            "0: [...] and to test that the whole thing works",
            "                                        ^",
        ],
        "line": 0,
        "column": 100,
    }


def test_text_buffer_insert():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.line = 1
    tb.column = 0

    tb.insert("XYZ\n")

    assert tb.current_line == "XYZ"
    assert tb.position == (1, 0)
    tb.nextline()
    assert tb.current_line == "def"
    assert tb.position == (2, 0)
